/* Copyright 2009 Patrick Higgins 
 *
 * This file is part of VladNput.
 *
 * VladNput is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * VladNput is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with VladNput.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vladnput;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * Aggregates the <tt>ParamValidator</tt> and <tt>ParamMatchValidator</tt>
 * instances for a single scope and allows for matching them in the proper
 * order efficiently.
 *
 * @author Patrick Higgins
 */
class ParamMatcher {
	
	// this class is immutable for thread safety, all instance variables must be final and
	// thread safe
	private final Map<String,ParamValidator> params;
	private final ParamMatchValidator[] paramMatches;

	/**
	 * Construct from a list of <tt>ParamValidator</tt> and <tt>ParamMatchValidator</tt> instances.
	 * The order of the <tt>param</tt> list is irrelevant. The <tt>paramMatch</tt> list should be
	 * in the order they appear in the source file.
	 * 
	 * @param param List of <tt>ParamValidator</tt>
	 * @param paramMatch List of <tt>ParamMatchValidator</tt>
	 */
	ParamMatcher(List<ParamValidator> param, List<ParamMatchValidator> paramMatch) {
		Map<String,ParamValidator> map = VladUtil.newHashMap();
		for (ParamValidator p : param) {
			map.put(p.getName(), p);
		}
		
		List<ParamMatchValidator> list = VladUtil.newArrayList();
		list.addAll(paramMatch);
		Collections.reverse(list);
		
		this.params = map.size() > 0 ? map : null;
		this.paramMatches = list.size() > 0 ?
				(ParamMatchValidator[]) list.toArray(new ParamMatchValidator[list.size()])
				: null;
	}
	
	/**
	 * Returns all the <tt>ParamValidator</tt> instances this instance contains,
	 * or <tt>null</tt> if there are none.
	 * 
	 * @return All <tt>ParamValidator</tt> instances, or <tt>null</tt> if there are none.
	 */
	Collection<ParamValidator> getParams() {
		if (params == null)
			return null;
		return params.values();
	}

	/**
	 * Adds all the <tt>ParamValidator</tt> and <tt>ParamMatchValidator</tt>
	 * that match <tt>param</tt> to <tt>list</tt> in proper precedence order.
	 * 
	 * @param param Parameter to match
	 * @param list List to add matching <tt>ParamValidator</tt> and <tt>ParamMatchValidtor</tt> instances to
	 */
	void addMatches(String param, List<BaseParamValidator> list) {
		if (params != null) {
			ParamValidator v = (ParamValidator) params.get(param);
			if (v != null) {
				list.add(v);
			}
		}
		for (int i = 0, len = paramMatches == null ? 0 : paramMatches.length; i < len; i++) {
			ParamMatchValidator v = paramMatches[i];
			if (v.getPattern().matcher(param).matches())
				list.add(v);
		}
	}

	StringBuffer toString(StringBuffer b, int indent) {
		if (params != null) {
			for (ParamValidator v : params.values()) {
				VladUtil.indent(b, indent).append(v.toString()).append("\n");
			}
		}
		for (int i = 0, len = paramMatches == null ? 0 : paramMatches.length; i < len; i++) {
			VladUtil.indent(b, indent).append(paramMatches[i].toString()).append("\n");
		}
		return b;
	}

}
