/* Copyright 2009 Patrick Higgins 
 *
 * This file is part of VladNput.
 *
 * VladNput is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * VladNput is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with VladNput.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vladnput;

import java.io.PrintStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.antlr.runtime.RecognitionException;

/**
 * Class for storing errors found during the parsing of a Vlad configuration
 * file.
 *
 * @author Patrick Higgins
 */
public class ParseErrors {

	private final List<String> errors = VladUtil.newArrayList();

	/**
	 * Add the error message from <tt>t.getMessage()</tt>.
	 * If <tt>t</tt> is a TokenException or RecognitionException, then source
	 * location information will be added automatically. Otherwise, the prefix
	 * "error: " will be added to the message.
	 * 
	 * @param t Throwable error
	 */
	public void add(Throwable t) {
		String msg = null;
		if (t instanceof TokenException) {
			TokenException e = (TokenException) t;
			msg = VladUtil.getSourceLocation(e) + " " + e.getMessage();
		}
		else if (t instanceof RecognitionException) {
			RecognitionException e = (RecognitionException) t;
			msg = VladUtil.getSourceLocation(e) + " " + e.getMessage();
		}
		else {
			msg = "error: " + t.getMessage();
		}
		errors.add(msg);
	}

	/**
	 * Add the error message.
	 * @param msg Error message
	 */
	public void add(String msg) {
		errors.add(msg);
	}

	/**
	 * Check if this object has any errors.
	 * @return <tt>true</tt> if there are errors, <tt>false</tt> otherwise.
	 */
	public boolean hasErrors() {
		return !errors.isEmpty();
	}

	/**
	 * Allows iterating over the errors.
	 * @return An iterator over all the errors.
	 */
	public Iterator<String> getErrors() {
		return errors.iterator();
	}
	
	private static interface Printer {
		void println(String s);
	}
	
	private void reportErrors(Printer out) {
		for (String msg : errors) {
			out.println(msg);
		}
	}
	
	public void reportErrors(final PrintWriter out) {
		reportErrors(new Printer() {
			public void println(String s) {
				out.println(s);
			}
		});
	}
	
	public void reportErrors(final PrintStream out) {
		reportErrors(new Printer() {
			public void println(String s) {
				out.println(s);
			}
		});
	}
	
}
