/* Copyright 2009 Patrick Higgins 
 *
 * This file is part of VladNput.
 *
 * VladNput is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * VladNput is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with VladNput.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vladnput;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import javax.servlet.http.HttpServletRequest;

import org.antlr.runtime.ANTLRFileStream;
import org.antlr.runtime.ANTLRInputStream;
import org.antlr.runtime.CharStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.TokenStream;
import org.antlr.runtime.tree.BufferedTreeNodeStream;
import org.antlr.runtime.tree.CommonTree;
import org.apache.log4j.Logger;

import org.vladnput.VladParser.start_return;
import org.vladnput.types.VladType;

/**
 * The main class for interfacing with the VladNput framework.
 * 
 * Obtain an instance using <tt>VladConfig.parse(file)</tt> or
 * <tt>VladConfig.parseResource(resourceName)</tt> and then
 * use <tt>vladConfig.validate(request, null)</tt> to validate HTTP
 * requests.
 *
 * This class is immutable and can be used safely from multiple threads
 * without synchronization. References to it may also be published
 * without synchronization on 1.5 JVMs and above.
 *
 * @author Patrick Higgins
 */
public class VladConfig {
	
	private static final Logger log = Logger.getLogger(VladConfig.class);

	/**
	 * The value that is put into <tt>VladError</tt> objects instead of the
	 * real value for parameters and types that are sensitive, e.g. passwords
	 * and credit cards.
	 */
	public static final String SENSITIVE_VALUE = "SENSITIVE";
	
	
	// this class is immutable for thread safety, all instance variables must be final and
	// thread safe
	private final ParamMatcher paramMatcher;
	private final URLMatcher urlMatcher;
	private final VladMessages defaultMsgs;
	private final VladLogger logger;
	private final Integer maxLog;
	private final Integer unrecognizedURLMaxLog;
	private final boolean enforce;
	private final boolean enforceUnrecognized;
	
	// code common to parse() and parseResource()
	private static VladConfig parse(CharStream stream, File srcFile) throws ParseException {
		ParseErrors errors = new ParseErrors();
		VladLexer lexer = new VladLexer(stream);
		lexer.setParseErrors(errors);
		TokenStream tokenStream = new CommonTokenStream(lexer);
		VladParser parser = new VladParser(tokenStream);
		parser.setParseErrors(errors);
		try {
			start_return ret = parser.start();
			CommonTree tree = (CommonTree) ret.tree;
			BufferedTreeNodeStream nodeStream = new BufferedTreeNodeStream(tree);
			nodeStream.setTokenStream(tokenStream);
			VladWalker walker = new VladWalker(nodeStream);
			walker.setParseErrors(errors);
			if (srcFile != null)
				walker.setSrcFile(srcFile);
			walker.start();
			
			VladConfig vladConfig = walker.buildVladConfig();
			if (vladConfig == null) {
				throw new ParseException(errors);
			}
			return vladConfig;
		}
		catch (RecognitionException e) {
			errors.add(e);
			throw new ParseException(errors);
		}
	}

	/**
	 * Parse a Vlad configuration from <tt>resourceName</tt>.
	 * 
	 * @param resourceName Vlad configuration resource
	 * @return Vlad configuration
	 * @throws IOException If there is a problem reading <tt>file</tt>
	 * @throws ParseException If there is a problem parsing <tt>file</tt>
	 */
	public static VladConfig parseResource(String resourceName) throws IOException, ParseException {
		final String key = VladUtil.resourceKey(resourceName);
		URL url = VladConfig.class.getClassLoader().getResource(resourceName);
		if (url == null)
			throw new FileNotFoundException(resourceName);
		File rel = VladUtil.relativize(url);
		return parse(new ANTLRInputStream(url.openStream()) {
			public String getSourceName() {
				return key;
			}
		}, rel);
	}

	/**
	 * Parse a Vlad configuration from <tt>file</tt>.
	 * 
	 * @param file Vlad configuration file
	 * @return Vlad configuration
	 * @throws IOException If there is a problem reading <tt>file</tt>
	 * @throws ParseException If there is a problem parsing <tt>file</tt>
	 */
	public static VladConfig parse(File file) throws IOException, ParseException {
		return parse(new ANTLRFileStream(file.getPath()), file);
	}

    VladConfig(List<ParamValidator> param,
    		List<ParamMatchValidator> paramMatch,
    		List<URLValidator> urls,
    		List<URLMatchValidator> urlMatches,
    		VladMessages defaultMsgs,
    		VladLogger logger,
    		Integer maxLog,
    		Integer unrecognizedURLMaxLog,
    		boolean enforce,
    		boolean enforceUnrecognized)
   	{
    	this.paramMatcher = new ParamMatcher(param, paramMatch);
    	this.urlMatcher = new URLMatcher(urls, urlMatches);
    	this.defaultMsgs = defaultMsgs;
    	this.logger = logger;
    	this.maxLog = maxLog;
    	this.unrecognizedURLMaxLog = unrecognizedURLMaxLog;
    	this.enforce = enforce;
    	this.enforceUnrecognized = enforceUnrecognized;
	}
    
    /**
     * Validates the <tt>request</tt> against this Vlad configuration.
     * If the request URL or parameter names require special semantics
     * such as case-insensitivity or an environment-specific prefix to
     * be stripped, then a <tt>canonicalizer</tt> should be supplied. If
     * one is not given, then the URL used will be <tt>request.getRequestURI()</tt>
     * and parameter names will be used as returned by <tt>request.getParameterNames()</tt>.
     * 
     * @param request The request to validate 
     * @param canonicalizer Canonicalizer for the request path and parameters. May be null.
     * @return All validation errors
     */
    public VladErrors validate(HttpServletRequest request, RequestCanonicalizer canonicalizer) {
    	String path = canonicalizer != null ? canonicalizer.getPath(request) : request.getRequestURI();
		VladMessageVars vars = new VladMessageVars(path);
    	VladErrors errors = new VladErrors();
    	List<BaseURLValidator> urlMatches = urlMatcher.match(path);

    	if (urlMatches.size() > 0) {

    		// validate the parameters we were given
        	for (Enumeration e = request.getParameterNames(); e.hasMoreElements(); ) {
        		String param = (String) e.nextElement();
        		String[] values = request.getParameterValues(param);
        		assert(values != null);

        		if (canonicalizer != null)
        			param = canonicalizer.canonicalizeParamName(param);

        		Scope scope = new Scope(param, urlMatches);
        		vars.put(VladMessageVars.PARAM_VAR, param);
        		vars.put(VladMessageVars.PARAM_DESC_VAR, scope.desc);
        		
        		if (scope.card == null) {
    				scope.addError(VladMessages.UNKNOWN_PARAM_MSG_KEY, vars, errors, request);
        		}
        		else {
                	// validate cardinality
        			switch (scope.card.test(values.length)) {
        			case Cardinality.VALID:
        				break;
        			case Cardinality.MISSING_REQUIRED:
        				scope.addError(VladMessages.REQ_MSG_KEY, vars, errors, request);
        				break;
        			case Cardinality.TOO_MANY:
        				scope.addError(VladMessages.TOO_MANY_MSG_KEY, vars, errors, request);
        				break;
        			case Cardinality.TOO_FEW:
        				scope.addError(VladMessages.TOO_FEW_MSG_KEY, vars, errors, request);
        				break;
        			default:
        				throw new IllegalStateException();
        			}

        			// validate data values
        			for (int i = 0; i < values.length; i++) {
        				List<String> msgs = scope.type.validate(values[i], vars);
        				if (msgs != null && msgs.size() > 0) {
        					if ((scope.sensitive != null && scope.sensitive.booleanValue()) || scope.type.isSensitive())
        						vars.put(VladMessageVars.VALUE_VAR, SENSITIVE_VALUE);
        					else
        						vars.put(VladMessageVars.VALUE_VAR, values[i]);
        					scope.addError(msgs, vars, errors, request);
        				}
        			}
        		}

        	}

        	// flag any required parameters we were not given
        	SortedSet<String> paramNames = new TreeSet<String>();
        	for (BaseURLValidator v : urlMatches) {
        		addParams(v.getParamMatcher(), paramNames);
        	}
        	addParams(paramMatcher, paramNames);

        	for (String param : paramNames) {
        		Scope scope = new Scope(param, urlMatches);
        		vars.put(VladMessageVars.PARAM_VAR, param);
        		vars.put(VladMessageVars.PARAM_DESC_VAR, scope.desc);
    			int testVal = scope.card.test(0);
    			if (testVal != Cardinality.VALID && request.getParameter(param) == null) {
            		vars.remove(VladMessageVars.VALUE_VAR);
    				switch (testVal) {
    				case Cardinality.TOO_FEW:
    					scope.addError(VladMessages.TOO_FEW_MSG_KEY, vars, errors, request);
    					break;
    				case Cardinality.MISSING_REQUIRED:
    					scope.addError(VladMessages.REQ_MSG_KEY, vars, errors, request);
    					break;
    				default:
    					throw new IllegalStateException();
    				}
    			}
        	}
    	}
    	else {
    		errors.add(new VladError(path, null, null, "Unrecognized URL", request.getRemoteAddr(), unrecognizedURLMaxLog, enforceUnrecognized));
    	}

    	logger.logErrors(errors);

    	return errors;
    }

    private void addParams(ParamMatcher matcher, Set<String> paramNames) {
    	Collection<ParamValidator> params = matcher.getParams();
    	if (params == null)
    		return;

    	for (ParamValidator pval : params) {
			paramNames.add(pval.getName());
    	}
    }

    private class Scope {
    	List<BaseParamValidator> paramMatches;
    	Cardinality card;
    	VladType type;
    	Integer scopeMaxLog;
    	String desc;
    	Boolean sensitive;
    	Boolean scopeEnforce;
    	
    	Scope(String param, List<BaseURLValidator> urlMatches) {
    		paramMatches = VladUtil.newArrayList();
    		for (BaseURLValidator v : urlMatches) {
        		v.getParamMatcher().addMatches(param, paramMatches);
        	}
        	paramMatcher.addMatches(param, paramMatches);

        	for (BaseParamValidator v : paramMatches) {
        		if (v.matches(param)) {
        			if (log.isDebugEnabled())
        				log.debug("Found match for param "+param+" in URLs "+urlMatches+": "+v);
            		if (card == null)
            			card = v.getCardinality();
            		if (type == null)
            			type = v.getType();
            		if (scopeMaxLog == null)
            			scopeMaxLog = v.getMaxLog();
            		if (desc == null)
            			desc = v.getDesc();
            		if (sensitive == null)
            			sensitive = v.getSensitive();
            		if (scopeEnforce == null)
            			scopeEnforce = v.getEnforce();
        		}
        	}

        	if (paramMatches.size() > 0) {
        		assert(card != null);
        		assert(type != null);
            	if (card == null) {
            		log.error("No cardinality found for "+param+": paramMatches="+paramMatches+", urlMatches="+urlMatches);
            	}
        	}
        	
    		if (desc == null)
    			desc = param;
    		
    		if (scopeMaxLog == null)
    			scopeMaxLog = maxLog; // maxLog comes from the outer VladConfig class 

    		if (scopeEnforce == null)
    			scopeEnforce = enforce; // enforce comes from the outer VladConfig class
    	}
    	
        public void addError(List<String> msgs, VladMessageVars vars, VladErrors errors, HttpServletRequest request) {
			VladError error = new VladError(vars.get(VladMessageVars.URI),
											vars.get(VladMessageVars.PARAM_VAR),
											vars.get(VladMessageVars.VALUE_VAR),
											getMsg(VladMessages.MSG_KEY, vars),
											request.getRemoteAddr(),
											scopeMaxLog,
											scopeEnforce);
			for (String msgKey : msgs) {
				String msg = getMsg(msgKey, vars);
				if (msg != null) {
					error.addDetail(getMsg(msgKey, vars));
				}
				else {
					if (log.isDebugEnabled())
						log.debug("Detail msg not found: "+msgKey);
				}
			}
			errors.add(error);
		}

		public void addError(String msgKey, VladMessageVars vars, VladErrors errors, HttpServletRequest request) {
			VladError error = new VladError(vars.get(VladMessageVars.URI),
											vars.get(VladMessageVars.PARAM_VAR),
											vars.get(VladMessageVars.VALUE_VAR),
											getMsg(msgKey, vars),
											request.getRemoteAddr(),
											scopeMaxLog,
											scopeEnforce);
			errors.add(error);
		}

		String getMsg(String msgKey, VladMessageVars vars) {
			String msg;
        	VladMessages msgs = null;
        	for (BaseParamValidator v : paramMatches) {
        		msgs = v.getMessages();
        		msg = msgs.getMessage(msgKey, vars);
        		if (msg != null)
        			return msg;
        		VladType type = v.getType();
        		if (type != null) {
            		msgs = v.getType().getMessages();
            		msg = msgs.getMessage(msgKey, vars);
            		if (msg != null)
            			return msg;
        		}
        	}
        	msgs = defaultMsgs;
    		return msgs.getMessage(msgKey, vars);
		}
		
    }

    public String toString() {
		StringBuffer b = new StringBuffer("VladConfig:").append("\n");
		paramMatcher.toString(b, 1);
		urlMatcher.toString(b, 1);
		b.append(":VladConfig");
		return b.toString();
	}

}
