/* Copyright 2009 Patrick Higgins 
 *
 * This file is part of VladNput.
 *
 * VladNput is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * VladNput is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with VladNput.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vladnput;

import java.util.ArrayList;
import java.util.List;

/**
 * Class representing a single validation error. It contains information about
 * the requesting IP address, URL, and parameter (if applicable) that failed
 * to validate, as well as a generic message and a list of detailed messages
 * about the error. There is also a maximum number of times that this URL and
 * parameter name combination should be logged and whether this error should
 * be enforced or not. Errors that are not enforced will be logged. It the
 * responsibility of the web application to reject requests that produce
 * enforced errors.
 * 
 * Detailed messages are helpful for the cases where a single type can validate
 * multiple aspects. For instance, a string type might check that the string
 * matches a regular expression as well as minimum and maximum length checks.
 * If a value violates both the regular expression and the maximum length, then
 * the detail messages can indicate that both constraints were violated while
 * the generic message can indicate that there was a problem with the
 * parameter without giving that level of detail. A user interface can present
 * these two levels of detail or choose to ignore one or the other as deemed
 * appropriate.
 * 
 * This class is not thread-safe.
 *
 * @author Patrick Higgins
 */
public class VladError {
	
	private final String url;
	private final String paramName;
	private final String paramValue;
	private final String message;
	private final String remoteAddr;
	private final Integer maxLog;
	private final boolean enforce;
	private List<String> details;
	
	/**
	 * Constructs a new error.
	 * 
	 * @param url The URL of the invalid request
	 * @param paramName The invalid parameter name, or <tt>null</tt> if the URL itself was the problem
	 * @param paramValue The invalid parameter value, or <tt>null</tt> if the URL itself was the problem
	 * @param message Generic message about the error
	 * @param remoteAddr The client IP address which issued the request
	 * @param maxLog The maximum number of times this url and parameter name combination should be logged, or <tt>null</tt> for no limit
	 * @param enforce Whether or not this error should be enforced
	 */
	public VladError(String url, String paramName, String paramValue, String message, String remoteAddr, Integer maxLog, boolean enforce) {
		this.url = url;
		this.paramName = paramName;
		this.paramValue = paramValue;
		this.message = message;
		this.remoteAddr = remoteAddr;
		this.maxLog = maxLog;
		this.enforce = enforce;
	}
	
	/**
	 * Returns the URL of the invalid request
	 * @return The URL
	 */
	public String getUrl() {
		return url;
	}

	/**
	 * Returns the name of the invalid parameter
	 * @return Parameter name
	 */
	public String getParamName() {
		return paramName;
	}

	/**
	 * Returns the value of the invalid parameter
	 * @return Parameter value
	 */
	public String getParamValue() {
		return paramValue;
	}

	/**
	 * Returns the generic error message
	 * @return Error message
	 */
	public String getMessage() {
		return message;
	}

	/**
	 * Returns the IP address of client that issued the invalid request
	 * @return Client IP address
	 */
	public String getRemoteAddr() {
		return remoteAddr;
	}

	/**
	 * Returns the maximum number of times this url and parameter name combination should be logged. 
	 * @return Maximum number of log entries, or <tt>null</tt> if there is no limit
	 */
	public Integer getMaxLog() {
		return maxLog;
	}
	
	/**
	 * Returns true if the error should be enforced (causing the request to be
	 * rejected), false otherwise.
	 * @return The enforcement value
	 */
	public boolean getEnforce() {
		return enforce;
	}

	/**
	 * Adds a detailed error message.
	 * @param msg The message
	 */
	public void addDetail(String msg) {
		if (details == null)
			details = VladUtil.newArrayList();
		details.add(msg);
	}

	/**
	 * Gets the detailed messages.
	 * @return List of String messages, or <tt>null</tt> if there are none.
	 */
	public List<String> getDetails() {
		return details;
	}

	public String toString() {
		return url + ": " + message;
	}
}
