/* Copyright 2009 Patrick Higgins 
 *
 * This file is part of VladNput.
 *
 * VladNput is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * VladNput is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with VladNput.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vladnput;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.atomic.AtomicInteger;

import org.apache.log4j.Logger;

/**
 * This class logs all requests and request parameters that are in violation
 * of the security policy defined by a Vlad configuration.
 * 
 * These errors are logged with Log4j and sent to the logger named
 * <tt>org.vladnput.VladLogger</tt> at a warn level.
 * 
 * The exact format of the log messages will depend on your local Log4j
 * configuration, but the log messages generated are in the following format:
 * 
 * <tt>[Client IP] [URL] [Param] [Param Value] [Generic Message] [Detail Message 1] [Detail Message 2] ...</tt>
 * 
 * There may not be any detail messages.
 * 
 * Instances of this class are thread safe.
 *
 * @author Patrick Higgins
 */
public class VladLogger {
	
	static final Logger log = Logger.getLogger(VladLogger.class);
	
	// maps LogKey instances to AtomicInteger classes. Is thread-safe because
	// ConcurrentHashMap and AtomicInteger are thread safe and LogKey is
	// immutable
	private final ConcurrentMap<LogKey,AtomicInteger> logCounts = new ConcurrentHashMap<LogKey,AtomicInteger>();

	// class for fast comparisons 
	static class LogKey {
		final String url;
		final String param;
		LogKey(String url, String param) {
			assert(url != null);
			this.url = url;
			this.param = param;
		}
		public int hashCode() {
			if (param == null)
				return url.hashCode();
			final int prime = 31;
			return prime * (prime + param.hashCode()) + url.hashCode();
		}
		public boolean equals(Object obj) {
			LogKey other = (LogKey) obj;
			return url.equals(other.url) &&
				(param == null && other.param == null ||
				 param != null && param.equals(other.param));
		}
		public String toString() {
			return url+" "+param;
		}
	}
	
	/**
	 * Logs <tt>errors</tt> to the Log4j logger. Does not log any errors that
	 * have already been logged <tt>VladError.getMaxLog()</tt> times.
	 * 
	 * @param errors The errors to log
	 */
	public void logErrors(VladErrors errors) {
		for (VladError error : errors) {
			Integer maxLogObj = error.getMaxLog();
			int maxLog = maxLogObj == null ? -1 : maxLogObj.intValue();
			if (maxLog == 0) {
				continue;
			}
			else if (maxLog > 0) {
				LogKey key = new LogKey(error.getUrl(), error.getParamName());
				AtomicInteger countVal = (AtomicInteger) logCounts.get(key);
				if (countVal == null) {
					countVal = new AtomicInteger(0);
					AtomicInteger oldVal = (AtomicInteger) logCounts.putIfAbsent(key, countVal);
					if (oldVal != null)
						countVal = oldVal;
				}
				int count = countVal.incrementAndGet();
				if (count < 0) { // overflow
					countVal.set(maxLog+1);
				}
				else if (count <= maxLog) {
					logError(error);					
				}
			}
			else {
				assert(maxLog < 0);
				logError(error);
			}
		}
	}
	
	private void logError(VladError error) {
		StringBuffer b = new StringBuffer("[");
		b.append(error.getRemoteAddr()).append("] [")
			.append(error.getUrl()).append("] [")
			.append(error.getParamName()).append("] [")
			.append(error.getParamValue()).append("] [")
			.append(error.getMessage()).append("]");

		List<String> details = error.getDetails();
		if (details != null) {
			for (String msg : details) {
				b.append(" [").append(msg).append("]");
			}
		}
		log.warn(b.toString());
	}

}
