/* Copyright 2009 Patrick Higgins 
 *
 * This file is part of VladNput.
 *
 * VladNput is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * VladNput is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with VladNput.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vladnput;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;

/**
 * Class that acts as a registry of all the variable names that may appear in
 * a message. The ${param}, ${param_desc}, ${value}, and ${uri} values are
 * always defined. More variables are defined by individual types.
 * 
 * Instances of this class hold the values of variables that are to be
 * interpolated into a message.
 *
 * @author Patrick Higgins
 */
public class VladMessageVars {

	public static final String PARAM_VAR = "param";
	public static final String PARAM_DESC_VAR = "param_desc";
	public static final String VALUE_VAR = "value";
	public static final String URI = "uri";

	private static final Set<String> VALID_VARS = VladUtil.newHashSet();
	static {
		VALID_VARS.add(PARAM_VAR);
		VALID_VARS.add(PARAM_DESC_VAR);
		VALID_VARS.add(VALUE_VAR);
		VALID_VARS.add(URI);
	}

	/**
	 * Registers the variable named <tt>var</tt> as a valid variable name that
	 * may appear inside a message body as <tt>${var}</tt>.
	 * @param var Variable name
	 */
	public static void registerVar(String var) {
		VALID_VARS.add(var);
	}

	/**
	 * A bulk version of <tt>registerVar(String)</tt> for a collection of
	 * String variable names.
	 * @param vars Variable names
	 */
	public static void registerVars(Collection<String> vars) {
		VALID_VARS.addAll(vars);
	}

	/**
	 * Returns true if <tt>key</tt> is a valid variable name, false otherwise.
	 * @param key Variable name
	 * @return true if <tt>key</tt> is a valid variable name, false otherwise
	 */
	public static boolean isValidVar(String key) {
		return VALID_VARS.contains(key);
	}

	private final Map<String,String> map = VladUtil.newHashMap();
	
	public VladMessageVars(String path) {
		map.put(URI, path);
	}

	/**
	 * Gets the value of the variable named <tt>var</tt>.
	 * @param var Message variable name
	 * @return Message variable value
	 */
	public String get(String var) {
		return map.get(var);
	}
	
	/**
	 * Sets the value of the variable named <tt>var</tt> to <tt>value</tt>.
	 * @param var Message variable name
	 * @param value Message variable value
	 */
	public void put(String var, String value) {
		assert(isValidVar(var));
		map.put(var, value);
	}

	/**
	 * Unsets the value of the variable named <tt>var</tt>.
	 * @param var Message variable name
	 */
	public void remove(String var) {
		map.remove(var);
	}

}
