/* Copyright 2009 Patrick Higgins 
 *
 * This file is part of VladNput.
 *
 * VladNput is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * VladNput is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with VladNput.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vladnput;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Class that defines valid message names and stores the messages for an
 * individual type, parameter, parameter match, or message defaults.
 *
 * @author Patrick Higgins
 */
public class VladMessages {

	/**
	 * Name of the generic message
	 */
	public static final String MSG_KEY = "msg";
	/**
	 * Name of the required message
	 */
	public static final String REQ_MSG_KEY = "req_msg";
	/**
	 * Name of the too many values given message
	 */
	public static final String TOO_MANY_MSG_KEY = "too_many_msg";
	/**
	 * Name of the too few values given message
	 */
	public static final String TOO_FEW_MSG_KEY = "too_few_msg";
	/**
	 * Name of the unknown parameter message
	 */
	public static final String UNKNOWN_PARAM_MSG_KEY = "unknown_param_msg";

	// this class is immutable for thread safety, all instance variables must be final and
	// thread safe
	private final Map<String,String> msgs = VladUtil.newHashMap();

	/**
	 * Set of valid message names.
	 */
	public static final Set<String> VALID_KEYS;
	static {
		Set<String> validKeys = VladUtil.newHashSet();

		validKeys.add(MSG_KEY);
		validKeys.add(REQ_MSG_KEY);
		validKeys.add(TOO_MANY_MSG_KEY);
		validKeys.add(TOO_FEW_MSG_KEY);
		validKeys.add(UNKNOWN_PARAM_MSG_KEY);

		VALID_KEYS = Collections.unmodifiableSet(validKeys);
	}
	
	/**
	 * Checks if <tt>name</tt> is a valid message name. Valid message names
	 * are either in <tt>VALID_KEYS</tt> or end with "_msg".
	 * @param name Message name
	 * @return <tt>true</tt> if <tt>name</tt> is a valid message name, <tt>false</tt> otherwise
	 */
	public static boolean isValidMessageName(String name) {
		return VALID_KEYS.contains(name) || name.endsWith("_msg");
	}

	/**
	 * Checks <tt>message</tt> to ensure that any variables references within
	 * it are well formed.
	 * @param message The message
	 * @return <tt>true</tt> if <tt>message</tt> is valid, <tt>false</tt> otherwise
	 */
	public static boolean isValidFormat(String message) {
        Matcher m = VladUtil.INTERPOLATE_REGEX.matcher(message);
        while (m.find()) {
            if (!VladMessageVars.isValidVar(m.group(1)))
            	return false;
        }
        return true;
	}

	/**
	 * Extracts all the arguments from <tt>args</tt> for which
	 * <tt>isValidMessageName()</tt> returns <tt>true</tt> and removes them
	 * from <tt>args</tt>.
	 * 
	 * Arguments that are not valid message names are ignored.
	 * 
	 * @param args Arguments
	 */
	@SuppressWarnings("unchecked")
	public VladMessages(Map args) {
		for (Iterator i = args.entrySet().iterator(); i.hasNext(); ) {
			Map.Entry entry = (Map.Entry) i.next();
			if (entry.getKey() instanceof String) {
				String key = (String) entry.getKey();
				if (isValidMessageName(key)) {
					i.remove();
					msgs.put(key, (String) entry.getValue());
				}
			}
		}
	}

	/**
	 * Returns the message named <tt>name</tt> after replacing any variables
	 * within it with the values from <tt>vars</tt>.
	 * @param name Message name
	 * @param vars Variable values
	 * @return Interpolated message, or <tt>null</tt> if the message does not exist
	 */
	public String getMessage(String name, VladMessageVars vars) {
		String msg = (String) msgs.get(name);
		if (msg == null)
			return null;
		return VladUtil.interpolate(msg, vars);
	}

}
