/* Copyright 2009 Patrick Higgins 
 *
 * This file is part of VladNput.
 *
 * VladNput is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * VladNput is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with VladNput.  If not, see <http://www.gnu.org/licenses/>.
 */
parser grammar VladParser;

options {
  tokenVocab=VladLexer;
  output=AST;
  ASTLabelType=CommonTree;
}

tokens {
  EMPTY_ARGS;
  KW_ARGS;
  NO_ARGS;
  NO_BODY;
  NO_CARD_SPEC;
  OPTION;
  QNAME;
  URL_BODY;
}

@header {
package org.vladnput;
}

@members {

private final static boolean DEBUG = false;

private ParseErrors parseErrors = null;

public void setParseErrors(ParseErrors parseErrors) {
    this.parseErrors = parseErrors;
}

public String getErrorHeader(RecognitionException e) {
    return VladUtil.getSourceLocation(e);
}

public void emitErrorMessage(String msg) {
    parseErrors.add(msg);
}

public String getErrorMessage(RecognitionException e, String[] tokenNames) {
    if (DEBUG) {
        List stack = getRuleInvocationStack(e, this.getClass().getName());
        String msg = null;
        if (e instanceof NoViableAltException) {
            NoViableAltException nvae = (NoViableAltException) e;
            msg = " no viable alt; token="+e.token+
                  " (decision="+nvae.decisionNumber+
                  " state "+nvae.stateNumber+")"+
                  " decision=<<"+nvae.grammarDecisionDescription+">>";
        }
        else {
            msg = super.getErrorMessage(e, tokenNames);
        }
        return stack+" "+msg;
    }
    else {
        return super.getErrorMessage(e, tokenNames);
    }
}
    
public String getTokenErrorDisplay(Token t) {
    return DEBUG ? t.toString() : super.getTokenErrorDisplay(t);
}

}

start
	:	stmt+ EOF
	;

stmt
	:	include_stmt SEMI!
	|	options_stmt
	|	type_stmt SEMI!
	|	param_stmt SEMI!
	|	param_match_stmt SEMI!
	|	url_stmt
	|	url_match_stmt
	;

include_stmt
	:	INCLUDE include_file
		-> ^(INCLUDE include_file)
	;
	
options_stmt
	:	OPTIONS LBRACE option+ RBRACE
		-> ^(OPTIONS option+)
	|	OPTIONS LBRACE RBRACE
		-> OPTIONS
	;

option
	:	ID EQUALS literal SEMI -> ^(OPTION ID literal)
	|	msg_defaults SEMI -> msg_defaults
	;

type_stmt
	:	TYPE ID EQUALS type_expr
		-> ^(TYPE ID type_expr)
	;
	
param_stmt
	:	PARAM param_id card_spec param_args
		-> ^(PARAM param_id card_spec param_args)
	;
	
param_id
	:	ID
	|	string
	;
	
param_match_stmt
	:	PARAM_MATCH string card_spec param_args
		-> ^(PARAM_MATCH string card_spec param_args)
	;

url_stmt
	:	URL url url_body
		-> ^(URL url url_body)
	;

url_match_stmt
	:	URL_MATCH string url_body
		-> ^(URL_MATCH string url_body)
	;

url_body
	:	LBRACE url_body_stmt+ RBRACE -> ^(URL_BODY url_body_stmt+)
	|	LBRACE RBRACE -> NO_BODY
	;
	
url_body_stmt
	:	param_stmt SEMI!
	|	param_match_stmt SEMI!
	;
	
include_file
	:	string
	|	ANGLE_STRING
	;

url
	:	URL_LITERAL
	|	string
	;

card_spec
	:	QUESTION
	|	STAR
	|	PLUS
	|	CARET
	|	LBRACE INT ( COMMA INT? )? RBRACE -> ^(LBRACE INT+)
	|	-> NO_CARD_SPEC // empty is valid
	;

param_args
	:	type_expr
	|	param_kw_args
	|	-> NO_ARGS // empty is valid
	;
	
type_expr
	:	qname type_kw_args?
	;

msg_defaults
	:	MESSAGE_DEFAULTS LPAREN msg_defaults_kw_args RPAREN
		-> ^(MESSAGE_DEFAULTS msg_defaults_kw_args)
	;

msg_defaults_kw_args
	:	msg_defaults_kw_arg ( COMMA msg_defaults_kw_arg )*
		-> ^(KW_ARGS msg_defaults_kw_arg+)
	;

msg_defaults_kw_arg
	:	ID EQUALS string -> ^(ID string)
	;

param_kw_args
	:	LPAREN param_kw_arg ( COMMA param_kw_arg )* RPAREN -> ^(KW_ARGS param_kw_arg+)
	|	LPAREN RPAREN -> EMPTY_ARGS
	;

param_kw_arg
	:	TYPE EQUALS type_expr -> ^(TYPE type_expr)
	|	ID EQUALS literal -> ^(ID literal)
	;

type_kw_args
	:	LPAREN type_kw_arg ( COMMA type_kw_arg )* RPAREN -> ^(KW_ARGS type_kw_arg+)
	|	LPAREN RPAREN -> EMPTY_ARGS
	;

type_kw_arg
	:	ID EQUALS literal -> ^(ID literal)
	;

qname
	:	ID ( DOT ID )+ -> ^(QNAME ID+)
	|	ID
	;
	
literal
	:	string
	|	INT
	|	FLOAT
	|	TRUE
	|	FALSE
	;
	
string
	:	DOUBLEQUOTED_STRING
	|	SINGLEQUOTED_STRING 
	;
