/* Copyright 2009 Patrick Higgins 
 *
 * This file is part of VladNput.
 *
 * VladNput is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * VladNput is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with VladNput.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vladnput;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

import org.antlr.runtime.ANTLRFileStream;
import org.antlr.runtime.ANTLRInputStream;
import org.antlr.runtime.CharStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.RecognizerSharedState;
import org.antlr.runtime.TokenStream;
import org.antlr.runtime.tree.CommonTree;
import org.antlr.runtime.tree.CommonTreeNodeStream;
import org.antlr.runtime.tree.TreeNodeStream;
import org.antlr.runtime.tree.TreeParser;
import org.apache.log4j.Logger;

import org.vladnput.VladParser.start_return;
import org.vladnput.types.InvalidArgException;
import org.vladnput.types.InvalidArgValueException;
import org.vladnput.types.MissingArgException;
import org.vladnput.types.VladType;

/**
 * Base class for the VladWalker tree grammar. This class handles most of the
 * semantic validation. It makes use of lots of inner classes that mirror
 * other classes in the org.vladnput package. The reason for this is that
 * we need mutable classes during parsing and semantic validation, but we want
 * immutable classes after the parsing and semantic validation are complete for
 * thread safety. The inner classes here are the mutable versions, and the
 * immutable versions are built when <tt>buildVladConfig()</tt> method is
 * called.
 *
 * @author Patrick Higgins
 */
class VladTreeParser extends TreeParser {
	
	protected static final Logger log = Logger.getLogger(VladTreeParser.class);
	
	private static class ParamHolder {
		final Map<String,Param> param = VladUtil.newLinkedHashMap();
		final Map<String,ParamMatch> paramMatch = VladUtil.newLinkedHashMap();
	}

	private static class ParseState {
		final Map<String,String> msgDefaults = VladUtil.newHashMap();
		final Map<String,Type> types = VladUtil.newHashMap();
		final Map<String,URL> urls = VladUtil.newLinkedHashMap();
		final Map<String,URLMatch> urlMatches = VladUtil.newLinkedHashMap();
		final ParamHolder globalParam = new ParamHolder();
		final List<String> included = VladUtil.newArrayList();
		ParseErrors errors = new ParseErrors();
		Integer maxLog = null;
		Integer unrecognizedURLMaxLog = null;
		Boolean enforce = null;
		Boolean enforceUnrecognized = null;
	}
	
	private static final Class<?>[] CONSTRUCTOR_ARGS = { Map.class };
	
	private ParseState parseState;

	protected BaseURL currentURL;
	protected BaseParam currentParam;
	private File directory;
	private VladLogger vladLogger = new VladLogger();

	// required constructor
    VladTreeParser(TreeNodeStream input, RecognizerSharedState state) {
		super(input, state);
		this.parseState = new ParseState();
		parseState.included.add(input.getSourceName());
	}

    // required constructor
    VladTreeParser(TreeNodeStream input) {
		super(input);
		this.parseState = new ParseState();
		parseState.included.add(input.getSourceName());
	}
    
    void setDirectory(File dir) {
    	if (!dir.isDirectory())
    		throw new IllegalArgumentException(dir + " is not a directory");
    	this.directory = dir;
    }

    void setSrcFile(File srcFile) {
    	if (!srcFile.isFile())
    		throw new IllegalArgumentException(srcFile + " is not a file");
    	File dir = srcFile.getParentFile();
    	if (dir == null)
    		dir = srcFile.getAbsoluteFile().getParentFile();
    	setDirectory(dir);
		parseState.included.add(srcFile.getPath());
    }
    
    void setLogger(VladLogger logger) {
    	this.vladLogger = logger;
    }
    
    void setErrors(ParseErrors errors) {
    	parseState.errors = errors;
    }

    ParseErrors getErrors() {
		return parseState.errors;
	}

    VladConfig buildVladConfig() {
		if (parseState.errors.hasErrors()) {
			log.warn("Not building VladConfig because of errors");
			return null;
		}
		
		ParamList globalParam = constructParams(parseState.globalParam);
		List<URLValidator> urls = constructURLs(parseState.urls);
		List<URLMatchValidator> urlMatches = constructURLs(parseState.urlMatches);
		VladMessages defaultMsgs = constructDefaultMessages();

		VladConfig vladConfig =
			new VladConfig(globalParam.param,
				globalParam.paramMatch,
				urls,
				urlMatches,
				defaultMsgs,
				vladLogger,
				parseState.maxLog,
				parseState.unrecognizedURLMaxLog,
				parseState.enforce == null ? true : parseState.enforce.booleanValue(),
				parseState.enforceUnrecognized == null ? true : parseState.enforceUnrecognized.booleanValue());

		if (parseState.errors.hasErrors()) {
			log.warn("Not returning VladConfig because of errors");
			return null;
		}
		return vladConfig;
	}
	
	private static class ParamList {
		final List<ParamValidator> param;
		final List<ParamMatchValidator> paramMatch;
		ParamList(List<ParamValidator> param, List<ParamMatchValidator> paramMatch) {
			this.param = param;
			this.paramMatch = paramMatch;
		}
	}
	
	@SuppressWarnings("unchecked")
	private <T extends BaseParam, U extends BaseParamValidator>
	List<U> constructParam(Map<String,T> map)
	{
		List<U> params = VladUtil.newArrayList();
		for (T param : map.values()) {
			try {
				// this cast is unchecked but necessary
				params.add((U) param.construct());
			}
			catch (ConstructionException e) {
				parseState.errors.add(e);
			}
		}
		return params;
	}

	private ParamList constructParams(ParamHolder param) {
		List<ParamValidator> pv = constructParam(param.param);
		List<ParamMatchValidator> pmv = constructParam(param.paramMatch);
		return new ParamList(pv, pmv);
	}

	@SuppressWarnings("unchecked")
	private <T extends BaseURL, U extends BaseURLValidator>
	List<U> constructURLs(Map<String,T> map)
	{
		List<U> urls = VladUtil.newArrayList();
		for (T url : map.values()) {
			ParamList params = constructParams(url);
			// this cast is unchecked but necessary
			urls.add((U) url.construct(params));
		}
		return urls;
	}

	private void setDefaultMessage(Map<String, Object> args, String key, String value) {
		if (!args.containsKey(key))
			args.put(key, value);
	}

	private VladMessages constructDefaultMessages() {
		Map<String,Object> args = VladUtil.newHashMap();
		args.putAll(parseState.msgDefaults);
		setDefaultMessage(args, VladMessages.MSG_KEY, "'${value}' is not a valid value for ${param_desc}");
		setDefaultMessage(args, VladMessages.REQ_MSG_KEY, "${param_desc} is required");
		setDefaultMessage(args, VladMessages.TOO_FEW_MSG_KEY, "Not enough values given for ${param_desc}");
		setDefaultMessage(args, VladMessages.TOO_MANY_MSG_KEY, "Too many values given for ${param_desc}");		
		setDefaultMessage(args, VladMessages.UNKNOWN_PARAM_MSG_KEY, "An unknown parameter was given: ${param}");
		return new VladMessages(args);
	}
	
	private void include(CharStream stream, File directory) throws IOException, RecognitionException {
		VladLexer lexer = new VladLexer(stream);
		lexer.setParseErrors(parseState.errors);
		TokenStream tokenStream = new CommonTokenStream(lexer);
		VladParser parser = new VladParser(tokenStream);
		parser.setParseErrors(parseState.errors);
		start_return ret = parser.start();
		CommonTree tree = (CommonTree) ret.tree;
		CommonTreeNodeStream nodeStream = new CommonTreeNodeStream(tree);
		nodeStream.setTokenStream(tokenStream);
		VladWalker walker = new VladWalker(nodeStream);
		if (directory != null)
			walker.setDirectory(directory);
		((VladTreeParser) walker).parseState = this.parseState;
		walker.start();
	}
	
	private void includeFile(File file, CommonTree node) {
		File rel = VladUtil.relativize(file);
		String path = rel.getPath();
		if (parseState.included.contains(path)) {
			parseState.errors.add(new RecursiveIncludeException(node.getText(), node));
		}
		else {
			parseState.included.add(path);
			try {
				include(new ANTLRFileStream(path), rel.getParentFile());
			}
			catch (RecognitionException e) {
				parseState.errors.add(e);
			}
			catch (FileNotFoundException e) {
				parseState.errors.add(new ResourceNotFoundException(node.getText(), node));
			}
			catch (IOException e) {
				parseState.errors.add(new ResourceException(node.getText(), node, e));
			}
		}
	}

	
	protected void includeFile(String filename, CommonTree node) {
		File file = new File(filename);
		if (file.isAbsolute()) {
			includeFile(file, node);			
		}
		else {
			includeFile(new File(directory, filename), node);
		}
	}
	
	protected void includeResource(CommonTree angleString) {
	   	String resourceName = VladUtil.angleString(angleString);
	   	final String resourceKey = VladUtil.resourceKey(resourceName);
	   	java.net.URL url = this.getClass().getClassLoader().getResource(resourceName);
	   	if (parseState.included.contains(resourceKey)) {
			parseState.errors.add(new RecursiveIncludeException(angleString.getText(), angleString));
	   	}
	   	else if (url == null) {
	   		parseState.errors.add(new ResourceNotFoundException(resourceName, angleString));
	   	}
	   	else {
	   		parseState.included.add(resourceKey);
	   		InputStream is = null;
	    	try {
		    	is = url.openStream();
	    		include(new ANTLRInputStream(is) {
	    			public String getSourceName() {
	    				return resourceKey;
	    			}
	    		}, null);
	    	}
			catch (RecognitionException e) {
				parseState.errors.add(e);
			}
	    	catch (IOException e) {
	    		parseState.errors.add(new ResourceException(resourceName, angleString, e));
	    	}
	    	finally {
	    		VladUtil.close(is);
	    	}
	   	}
	}
	
	protected void addOption(CommonTree idNode, Object value, CommonTree valueNode) {
		String option = idNode.getText();
		if ("unrecognized_url_max_log".equals(option)) {
			if (value instanceof Integer)
				parseState.unrecognizedURLMaxLog = (Integer) value;
			else
				parseState.errors.add(new InvalidKeywordValueException(valueNode, value));
		}
		else if ("max_log".equals(option)) {
			if (value instanceof Integer)
				parseState.maxLog = (Integer) value;
			else
				parseState.errors.add(new InvalidKeywordValueException(valueNode, value));			
		}
		else if ("enforce".equals(option)) {
			if (value instanceof Boolean)
				parseState.enforce = (Boolean) value;
			else
				parseState.errors.add(new InvalidKeywordValueException(valueNode, value));
		}
		else if ("enforce_unrecognized".equals(option)) {
			if (value instanceof Boolean)
				parseState.enforceUnrecognized = (Boolean) value;
			else
				parseState.errors.add(new InvalidKeywordValueException(valueNode, value));
		}
		else {
			parseState.errors.add(new InvalidKeywordException(idNode));
		}
	}
	
	protected void addMessageDefault(CommonTree node, String value) {
		String key = node.getText();
		if (VladMessages.isValidMessageName(key)) {
			if (VladMessages.isValidFormat(value))
				parseState.msgDefaults.put(key, value);
			else
				parseState.errors.add(new InvalidKeywordValueException(node, value));
		}
		else {
			parseState.errors.add(new InvalidKeywordException(node));
		}
	}
	
	protected void defineType(CommonTree name, Type type) {
		String key = name.getText();
		if (parseState.types.containsKey(key)) {
			Type orig = (Type) parseState.types.get(name);
			parseState.errors.add(new DuplicateException("type", name.token, orig.nameNode.token));
		}
		else {
			log.debug("define type: " + key + ": " + type);
			parseState.types.put(key, type);
		}
	}
	
	private boolean checkDupe(String key, CommonTree node, Map<String,? extends BaseURL> map, String msg) {
		if (map.containsKey(key)) {
			BaseURL old = map.get(key);
			parseState.errors.add(new DuplicateException(msg, node.token, old.node.token));
			return true;
		}
		return false;
	}
	
	protected URL addURL(String name, CommonTree node) {
		final String KEYWORD = "url";
		URL url = new URL(name, node);
		String key = name;
		if (!checkDupe(key, node, parseState.urlMatches, KEYWORD) &&
			!checkDupe(key, node, parseState.urls, KEYWORD)) {
			parseState.urls.put(key, url);
		}
		return url;
	}
	
	protected URLMatch addURLMatch(Pattern pattern, CommonTree node) {
		final String KEYWORD = "url_match";
		URLMatch url = new URLMatch(pattern, node);
		String key = Pattern.quote(pattern.pattern());
		if (!checkDupe(key, node, parseState.urlMatches, KEYWORD) &&
			!checkDupe(key, node, parseState.urls, KEYWORD)) {
			parseState.urlMatches.put(key, url);
		}
		return url;
	}
	
	private ParamHolder currentParam() {
		return currentURL == null ? parseState.globalParam : currentURL;
	}

	private boolean checkParamDupe(String key, CommonTree node, Map<String,? extends BaseParam> map, String msg) {
		if (map.containsKey(key)) {
			BaseParam old = map.get(key);
			parseState.errors.add(new DuplicateException(msg, node.token, old.node.token));
			return true;
		}
		return false;
	}
	
	protected Param addParam(String name, CommonTree node) {
		final String KEYWORD = "param";
		Param param = new Param(name, node);
		String key = name;
		ParamHolder holder = currentParam();
		if (!checkParamDupe(key, node, holder.param, KEYWORD) &&
			!checkParamDupe(key, node, holder.paramMatch, KEYWORD)) {
			holder.param.put(key, param);
		}
		else {
			param.ignore = true;
		}
		return param;
	}

	protected ParamMatch addParamMatch(Pattern pattern, CommonTree node) {
		final String KEYWORD = "param_match";
		ParamMatch param = new ParamMatch(pattern, node);
		String key = Pattern.quote(pattern.pattern());
		ParamHolder holder = currentParam();
		if (!checkParamDupe(key, node, holder.param, KEYWORD) &&
			!checkParamDupe(key, node, holder.paramMatch, KEYWORD)) {
			holder.paramMatch.put(key, param);
		}
		else {
			param.ignore = true;
		}
		return param;
	}

	private List<URLMatch> urlPrecedenceList(URL url) {
		List<URLMatch> list = new ArrayList<URLMatch>();
		for (URLMatch match : parseState.urlMatches.values()) {
			if (match.pattern.matcher(url.url).matches()) {
				list.add(match);
			}
		}
		Collections.reverse(list);
		return list;
	}
	
	private void addPrecedenceParams(Param param, List<BaseParam> list, Map<String,Param> map) {
		int matchCount = 0;
		for (Param p : map.values()) {
			if (param.name.equals(p.name)) {
				list.add(p);
				assert(++matchCount == 1); // there should be at most one match
			}
		}
	}

	private void addPrecedenceMatches(Param param, List<BaseParam> list, Map<String,ParamMatch> map) {
		// we can have more than one match, but we have to add them to list in reverse order
		// to obey the precedence rule that later param_match statements have precedence
		List<ParamMatch> matches = new ArrayList<ParamMatch>();
		for (ParamMatch p : map.values()) {
			if (p.pattern.matcher(param.name).matches()) {
				matches.add(p);
			}
		}
		Collections.reverse(matches);
		list.addAll(matches);
	}
	
	private List<BaseParam> paramPrecedenceList(Param param) {
		List<BaseParam> list = new ArrayList<BaseParam>();
		if (currentURL != null) {
			// param in url is this one
			addPrecedenceMatches(param, list, currentURL.paramMatch); // param_match in url
			if (currentURL instanceof URL) {
				List<URLMatch> matchingURLs = urlPrecedenceList((URL) currentURL);
				for (Iterator<URLMatch> i = matchingURLs.iterator(); i.hasNext(); ) {
					URLMatch urlMatch = i.next();
					addPrecedenceParams(param, list, urlMatch.param); // param in url_match
					addPrecedenceMatches(param, list, urlMatch.paramMatch); // param_match in url_match
				}
			}
			addPrecedenceParams(param, list, parseState.globalParam.param); // param in top-level
		}
		addPrecedenceMatches(param, list, parseState.globalParam.paramMatch); // param_match in top-level
		if (log.isDebugEnabled())
			log.debug("Precedence list for: "+param+": "+list);
		return list;
	}
	
	protected void finishParam(Param param) {
		if (param.ignore)
			return;

		List<BaseParam> precedence = paramPrecedenceList(param);

		boolean cardinalityFound = param.cardinality != null;
		if (!cardinalityFound) {
			for (Iterator<BaseParam> i = precedence.iterator(); i.hasNext(); ) {
				BaseParam p = i.next();
				if (p.cardinality != null)
					cardinalityFound = true;
			}
		}
		if (!cardinalityFound)
			parseState.errors.add(new MissingInheritedValueException("cardinality", param.node, precedence));

		boolean typeFound = param.type != null;
		if (!typeFound) {
			for (Iterator<BaseParam> i = precedence.iterator(); i.hasNext(); ) {
				BaseParam p = i.next();
				if (p.type != null)
					typeFound = true;
			}
		}
		if (!typeFound)
			parseState.errors.add(new MissingInheritedValueException("type", param.node, precedence));
	}

	protected void finishParamMatch(ParamMatch param) {
		if (param.ignore)
			return;

		// nothing to do
	}

	protected void addParamKwArg(CommonTree node, Object value) {
		try {
			currentParam.addArg(node, value);
		}
		catch (InvalidKeywordValueException e) {
			log.debug(e);
			parseState.errors.add(e);
		}		
		catch (InvalidKeywordException e) {
			log.debug(e);
			parseState.errors.add(e);
		}		
		catch (ClassCastException e) {
			log.debug(e);
			parseState.errors.add(new InvalidKeywordValueException(node, value));
		}
	}
	
	protected void addTypeArg(Type type, CommonTree node, Object value) {
		type.addArg(node, value);
	}
	
	protected void typeArgsComplete(Type type) {
		type.checkArgs(parseState.errors);
	}
	
	private static final Pattern INVALID_PATTERN = Pattern.compile("invalid_pattern");

	protected Pattern compileRegex(String pattern, CommonTree node) {
	    try {
	        return Pattern.compile(pattern);
	    }
	    catch (PatternSyntaxException e) {
	    	parseState.errors.add(new BadRegexException(node, e));
	    	return INVALID_PATTERN;
	    } 
	}
	
	protected String parseString(CommonTree node) {
		return VladUtil.string(node);
	}
	
	protected Double parseFloat(CommonTree node) {
		try {
			return Double.valueOf(node.getText());
		}
		catch (NumberFormatException e) {
			parseState.errors.add(new BadNumericException(node, e));
			return Double.valueOf(Double.NaN);
		}
	}
	
	protected Integer parseInt(CommonTree node) {
		try {
			return Integer.valueOf(node.getText());
		}
		catch (NumberFormatException e) {
			parseState.errors.add(new BadNumericException(node, e));
			return Integer.valueOf(Integer.MIN_VALUE);
		}
	}
	
	@SuppressWarnings("unchecked")
	protected Class<? extends VladType> parseClass(List<CommonTree> idNodes) {
		CommonTree first = idNodes.get(0);
		CommonTree last = idNodes.get(idNodes.size()-1);
		int tokensBack = first.token.getTokenIndex() - last.token.getTokenIndex();
		String className = input.getTokenStream().toString(first.getTokenStartIndex(), last.getTokenStopIndex());
		try {
			Class<?> klass = Class.forName(className);
			if (!VladType.class.isAssignableFrom(klass)) {
				parseState.errors.add(new BadTypeException(input, className, tokensBack));
				return null;
			}
			return (Class<? extends VladType>) klass;
		}
		catch (ClassNotFoundException e) {
			first.token.getTokenIndex();
			parseState.errors.add(new MissingClassException(input, className, tokensBack));
			return null;
		}
	}
	
	protected Type resolveType(CommonTree node) {
		String name = node.getText();
		Type type = (Type) parseState.types.get(name);
		if (type == null || type.getClass() == null) {
			parseState.errors.add(new UndefinedTypeException(name, node));
			return null;
		}
		log.debug("resolved type: " + name + " to " + type);
		return type;
	}
	
	protected void setCardinality(BaseParam param, CommonTree minNode, CommonTree maxNode)
		throws RecognitionException
	{
		IllegalCardinalityException ice = null;
		int min = parseInt(minNode).intValue();
		if (maxNode != null) {
			final int MIN_TOKEN = -3;
			final int MAX_TOKEN = -1;
			int max = parseInt(maxNode).intValue();
			if (min < 0)
				ice = new IllegalCardinalityException("negative min: " + min, input, MIN_TOKEN);
			else if (max < 0)
				ice = new IllegalCardinalityException("negative max: " + max, input, MAX_TOKEN);
			else if (min > max)
				ice = new IllegalCardinalityException("min ("+min+") greater than max ("+max+")", input, MIN_TOKEN);
			else
				param.setCardinality(min, max);
		}
		else {
			final int MIN_TOKEN = -1;
			if (min < 0)
				ice = new IllegalCardinalityException("negative min: " + min, input, MIN_TOKEN);
			else
				param.setCardinality(min);
		}
		if (ice != null) {
			parseState.errors.add(ice);
			 // set it to prevent further errors about cardinality inheritance
			param.setCardinality(Cardinality.REQUIRED);
		}
	}

	/************************** Types **********************/
	
	abstract static class BaseURL extends ParamHolder {
		final CommonTree node;

		BaseURL(CommonTree node) {
			this.node = node;
		}
		
		abstract BaseURLValidator construct(ParamList params);
	}
	
	private static class URL extends BaseURL {
		private final String url;

		URL(String url, CommonTree node) {
			super(node);
			this.url = url;
		}

		URLValidator construct(ParamList params) {
			return new URLValidator(url, params.param, params.paramMatch);
		}
	}

	private static class URLMatch extends BaseURL {
		private final Pattern pattern;

		URLMatch(Pattern pattern, CommonTree node) {
			super(node);
			this.pattern = pattern;
		}		

		URLMatchValidator construct(ParamList params) {
			return new URLMatchValidator(pattern, params.param, params.paramMatch);
		}
	}

	abstract static class BaseParam {
		final CommonTree node; // needed for line info on duplicate declarations
		boolean ignore = false;
		String desc;
		Cardinality cardinality;
		Type type;
		Integer maxLog;
		Boolean sensitive;
		Boolean enforce;
		final Map<String,Object> args = VladUtil.newHashMap();
		
		BaseParam(CommonTree node) {
			assert(node.token.getInputStream() != null);
			this.node = node;
		}
		
		abstract BaseParamValidator construct() throws ConstructionException;
		
        void addArg(CommonTree node, Object value) throws ClassCastException, InvalidKeywordException, InvalidKeywordValueException {
        	String name = node.getText();
        	if ("max_log".equals(name)) {
        		maxLog = (Integer) value;
        	}
        	else if ("desc".equals(name)) {
        		this.desc = (String) value;
        	}
        	else if ("enforce".equals(name)) {
        		this.enforce = (Boolean) value;
        	}
        	else if (VladType.SENSITIVE_KEY.equals(name)) {
        		this.sensitive = (Boolean) value;
        	}
        	else if (VladMessages.isValidMessageName(name)) {
        		if (VladMessages.isValidFormat((String) value))
        			args.put(name, value);
        		else
        			throw new InvalidKeywordValueException(node, value);
        	}
        	else {
        		throw new InvalidKeywordException(node);
        	}
		}

        void setCardinality(Cardinality checker) {
			this.cardinality = checker;
		}

        void setType(Type type) {
			this.type = type;
		}
		
        void setCardinality(int min) {
			if (min < 0)
				throw new IllegalArgumentException();
			switch (min) {
			case 0:	this.cardinality = Cardinality.ZERO_OR_MORE;	break;
			case 1:	this.cardinality = Cardinality.ONE_OR_MORE;	break;
			default:this.cardinality = new Cardinality.MinChecker(min);
			}
		}
		
        void setCardinality(int min, int max) {
			if (min < 0 || max < 0 || min > max)
				throw new IllegalArgumentException();

			if (min == 0 && max == 1)
				this.cardinality = Cardinality.OPTIONAL;
			else if (min == 1 && max == 1)
				this.cardinality = Cardinality.REQUIRED;
			else
				this.cardinality = new Cardinality.MinMaxChecker(min, max);
		}
		
		private Map<String,Object> mergeArgs() {
			List<Map<String, Object>> argList = new ArrayList<Map<String, Object>>();
			argList.add(args);
			for (Type parent = type; parent != null; parent = parent.parentType) {
				argList.add(parent.args);
			}			

			Map<String,Object> merged = VladUtil.newHashMap();
			for (int i = argList.size()-1; i >= 0; i--) {
				merged.putAll(argList.get(i));
			}
			return merged;
		}
		
		protected VladType constructType() throws ConstructionException {
			if (type == null)
				return null;

			Class<? extends VladType> klass = type.typeClass;
			try {
				Constructor<? extends VladType> ctor = klass.getConstructor(CONSTRUCTOR_ARGS);
				Map<String,Object> merged = mergeArgs();
				VladType vladType = (VladType) ctor.newInstance(new Object[] { merged });
				if (merged.size() > 0) {
					throw new ConstructionException("Unknown keys: " + merged.keySet(), klass, node, null);
				}
				return vladType;
			}
			catch (ConstructionException e) {
				throw e;
			}
			catch (NoSuchMethodException e) {
				throw new ConstructionException("missing Map constructor", klass, node, e);
			}
			catch (Exception e) {
				throw new ConstructionException(e.getMessage(), klass, node, e);
			}
		}
		
		public String toString() {
			return ""+type+", "+cardinality;
		}

	}

	static class Param extends BaseParam {
		final String name;

		Param(String name, CommonTree node) {
			super(node);
			this.name = name;
		}
		
		ParamValidator construct() throws ConstructionException {
        	VladMessages msgs = new VladMessages(args);
			return new ParamValidator(name, desc, cardinality, maxLog, constructType(), msgs, sensitive, enforce);
		}

		public String toString() {
			return "[param: "+name+" "+super.toString()+"]";
		}
	}

	static class ParamMatch extends BaseParam {
		final Pattern pattern;

		ParamMatch(Pattern pattern, CommonTree node) {
			super(node);
			this.pattern = pattern;
		}
		
		ParamMatchValidator construct() throws ConstructionException {
        	VladMessages msgs = new VladMessages(args);
			return new ParamMatchValidator(pattern, cardinality, maxLog, desc, constructType(), msgs, sensitive, enforce);
		}
		
		public String toString() {
			return "[param_match: "+pattern+" "+super.toString()+"]";
		}
	}

	static class Type {
		final CommonTree nameNode; // needed for line info on duplicate declarations
		final String typeName;
		final Class<? extends VladType> typeClass;
		final Type parentType;
		final Map<String,Object> args = VladUtil.newLinkedHashMap();
		final Map<String,CommonTree> argNodes = VladUtil.newHashMap();

		Type(CommonTree nameNode, Class<? extends VladType> typeClass) {
			this.nameNode = nameNode;
			this.typeClass = typeClass;
			this.parentType = null;
			this.typeName = typeClass != null ? typeClass.getName() : nameNode.getText();
		}

		Type(CommonTree nameNode, Type parentType) {
			this.nameNode = nameNode;
			this.typeName = nameNode.getText();
			if (parentType == null) {
				this.parentType = null;
				this.typeClass = null;
			}
			else {
				this.parentType = parentType;
				this.typeClass = parentType.typeClass;
				args.putAll(parentType.args);
			}
		}

        void addArg(CommonTree node, Object value) {
        	String name = node.getText();
        	args.put(name, value);
        	argNodes.put(name, node);
		}
        
		void checkArgs(ParseErrors errors) {
			Method checkArg = null;
			Method checkArgs = null;
			if (typeClass != null) {
				try {
					checkArg = typeClass.getMethod("checkArg", new Class[] { String.class, Object.class, Map.class });
				} catch (SecurityException e) {
					throw new IllegalStateException(e);
				} catch (NoSuchMethodException e) {
					throw new IllegalStateException(e);
				}

				try {
					checkArgs = typeClass.getMethod("checkArgs", new Class[] { Map.class });
				} catch (SecurityException e) {
					throw new IllegalStateException(e);
				} catch (NoSuchMethodException e) {
				}
			}
			
			Map<String, Object> constArgs = Collections.unmodifiableMap(args);

			for (Iterator<Map.Entry<String,Object>> i = args.entrySet().iterator(); i.hasNext(); ) {
				Map.Entry<String,Object> entry = i.next();
				String name = entry.getKey();
				Object value = entry.getValue();
				CommonTree node = argNodes.get(name);

				if (checkArg != null) {
					try {
						checkArg.invoke(null, new Object[] { name, value, constArgs });
					}
					catch (InvocationTargetException e) {
						Throwable cause =  e.getCause();
						if (cause instanceof ClassCastException) {
							errors.add(new InvalidKeywordValueException(node, value));
						}
						else if (cause instanceof InvalidArgValueException) {
							errors.add(new InvalidKeywordValueException(node, value));
						}
						else if (cause instanceof InvalidArgException) {
							errors.add(new InvalidKeywordException(node));
						}
						else {
							throw new IllegalStateException(cause);
						}
					}
					catch (IllegalAccessException e) {
						throw new IllegalStateException(e);
					}
				}
			}
			
			if (checkArgs != null) {
				try {
					checkArgs.invoke(null, new Object[] { constArgs });
				}
				catch (InvocationTargetException e) {
					Throwable cause =  e.getCause();
					if (cause instanceof MissingArgException) {
						MissingArgException mae = (MissingArgException) cause;
						errors.add(new MissingRequiredArgumentException(nameNode, typeName, mae.getName()));
					}
					else {
						throw new IllegalStateException(cause);
					}
				}
				catch (IllegalAccessException e) {
					throw new IllegalStateException(e);
				}				
			}
		}

		public String toString() {
			return "[type: "+typeClass+"; args: "+args+"]";
		}
	}
	
}
