/* Copyright 2009 Patrick Higgins 
 *
 * This file is part of VladNput.
 *
 * VladNput is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * VladNput is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with VladNput.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vladnput;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URL;
import java.text.DateFormat;
import java.text.ParsePosition;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.antlr.runtime.IntStream;
import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.Token;
import org.antlr.runtime.tree.CommonTree;
import org.apache.log4j.Logger;

/**
 * Various static utility routines.
 *
 * @author Patrick Higgins
 */
public class VladUtil {
	
	private static Logger log = Logger.getLogger(VladUtil.class);
	
	// prevent construction
	private VladUtil() { }

	private static String removeDelimiters(Token tok) {
		String text = tok.getText().substring(1);
		return text.substring(0, text.length()-1);
	}

    static String angleString(CommonTree node) {
		return removeDelimiters(node.token);
	}

    static File relativize(File file) {
    	try {
    		URI uri = file.getCanonicalFile().toURI();
    		URI cwd = new File(System.getProperty("user.dir")).getCanonicalFile().toURI();
    		URI rel = cwd.relativize(uri);
    		return new File(rel.getPath());
    	}
    	catch (IOException e) {
    		throw new RuntimeException(e);
    	}
    }
    
    static String resourceKey(String resourceName) {
    	return "resource:"+resourceName;
    }

    static File relativize(URL url) {
		if ("file".equals(url.getProtocol())) {
			return relativize(new File(url.getPath()));
		}
		return null;
    }
	
    static String string(CommonTree node) {
		Token tok = node.token;
		switch (tok.getType()) {
		case VladWalker.SINGLEQUOTED_STRING:
			return removeDelimiters(tok);
		case VladWalker.DOUBLEQUOTED_STRING:
			return parseStringLiteral(tok);
		case VladWalker.URL_LITERAL:
		case VladWalker.ID:
			return tok.getText();			
		default:
			throw new IllegalStateException(node.token.toString());
		}
	}
	
    private static String parseStringLiteral(Token tok) {
    	String str = tok.getText();
        char[] strArr = str.toCharArray();
        
        if (strArr[0] != '"' || strArr[strArr.length-1] != '"') {
        	// this is not a string literal, probably already parsed...
        	return str;
        }
        
        boolean escape = false;
        StringBuffer buf = new StringBuffer();
        for (int i = 1, len = strArr.length - 1; i < len; ++i) {
            if (escape) {
                if (strArr[i] == 'b') {
                    buf.append('\b');
                } else if (strArr[i] == 't') {
                    buf.append('\t');
                } else if (strArr[i] == 'n') {
                    buf.append('\n');
                } else if (strArr[i] == 'r') {
                    buf.append('\r');
                } else if (strArr[i] == 'f') {
                    buf.append('\f');
                } else if (strArr[i] == 'u') {
                    // Unicode escape
                    int utf = Integer.parseInt(str.substring(i + 1, i + 5), 16);
                    buf.append((char)utf);
                    i += 4;
                } else if (isOctalDigit(strArr[i])) {
                    // Octal escape
                    int j = 0;
                    for (j = 1; (j < 3) && (i + j < strArr.length); ++j) {
                        if (!isOctalDigit(strArr[i+j]))
                            break;
                    }
                    if (j == 3 && strArr[i] > '3') {
                        j = 2;
                    }
                    int octal = Integer.parseInt(str.substring(i, i + j), 8);
                    buf.append((char)octal);
                    i += j-1;
                } else {
                    buf.append(strArr[i]);
                }
                escape = false;
            } else if (strArr[i] == '\\') {
                escape = true;
            } else {
                buf.append(strArr[i]);
            }
        }
        return buf.toString();
    }
    
    private static boolean isOctalDigit(char c) {
        return (c >= '0' && c <= '7');
    }
	
    static void close(InputStream is) {
		if (is != null) {
			try {
				is.close();
			}
			catch (IOException e) {
				throw new RuntimeException(e);
			}
		}
	}
	
	private static String getSourceLocation(IntStream input, int line, int charPositionInLine) {
		assert(input != null);
	    return input.getSourceName()+":"+line+":"+(charPositionInLine);
	}
	
    static String getSourceLocation(RecognitionException e) {
		assert(e != null);
		return getSourceLocation(e.input, e.line, e.charPositionInLine);
	}

    static String getSourceLocation(Token t) {
		assert(t != null);
		assert(t.getInputStream() != null);
		return getSourceLocation(t.getInputStream(), t.getLine(), t.getCharPositionInLine());
	}

    static String getSourceLocation(TokenException e) {
		assert(e != null);
		assert(e.getToken() != null);
		return getSourceLocation(e.getToken());
	}
    
    static StringBuffer indent(StringBuffer b, int level) {
    	for (int i = 0; i < level; i++)
    		b.append("\t");
    	return b;
    }
    
	/**
	 * Removes and returns the mapping associated with <tt>key</tt> in
	 * <tt>args</tt>, or <tt>null</tt> if there is not one.
	 * @param args Arguments
	 * @param key Argument name
	 * @return Argument value, or <tt>null</tt> if there is not one.
	 */
    public static String stringKey(Map<String,Object> args, String key) {
		return (String) args.remove(key);
	}
	
	/**
	 * Removes and returns the mapping associated with <tt>key</tt> in
	 * <tt>args</tt>, or <tt>null</tt> if there is not one.
	 * @param args Arguments
	 * @param key Argument name
	 * @return Argument value, or <tt>null</tt> if there is not one.
	 */
	public static Integer integerKey(Map<String,Object> args, String key) {
		return (Integer) args.remove(key);
	}

	/**
	 * Removes and returns the mapping associated with <tt>key</tt> in
	 * <tt>args</tt>, or <tt>null</tt> if there is not one.
	 * @param args Arguments
	 * @param key Argument name
	 * @return Argument value, or <tt>null</tt> if there is not one.
	 */
	public static Double floatKey(Map<String,Object> args, String key) {
		return (Double) args.remove(key);
	}

	/**
	 * Removes and returns the mapping associated with <tt>key</tt> in
	 * <tt>args</tt>, or <tt>null</tt> if there is not one.
	 * @param args Arguments
	 * @param key Argument name
	 * @return Argument value, or <tt>null</tt> if there is not one.
	 */
	public static Boolean booleanKey(Map<String,Object> args, String key) {
		return (Boolean) args.remove(key);
	}

	/**
	 * Removes and returns the string mapping associated with <tt>key</tt> in
	 * <tt>args</tt> after parsing it as a date, or <tt>null</tt> if there is
	 * not one.
	 * @param args Arguments
	 * @param key Argument name
	 * @param format Date format
	 * @return Argument value, or <tt>null</tt> if there is not one.
	 * @throws IllegalArgumentException If the date is invalid
	 */
	public static Date dateKey(Map<String,Object> args, String key, DateFormat format) throws IllegalArgumentException {
		return parseDate(stringKey(args, key), format);
	}

	/**
	 * Parses <tt>value</tt> as a date using <tt>format</tt>. Returns <tt>null</tt>
	 * if <tt>value</tt> is <tt>null</tt>. Uses <tt>DateFormat.getDateInstance()</tt>
	 * if <tt>format</tt> is <tt>null</tt>.
	 * 
	 * @param value Value to parse
	 * @param format Date format
	 * @return Parsed value, or <tt>null</tt> if <tt>value</tt> is <tt>null</tt>
	 * @throws IllegalArgumentException If the date is invalid
	 */
	public static Date parseDate(String value, DateFormat format) throws IllegalArgumentException {
		if (value == null)
			return null;
		
		if (format == null)
			format = DateFormat.getDateInstance();
		ParsePosition pos = new ParsePosition(0);
		Date d = format.parse(value, pos);
		if (d == null || pos.getIndex() != value.length()) {
			throw new IllegalArgumentException("Failed to parse date: "+value);
		}
		return d;
	}

	/**
	 * The regular expression defining what variables in message bodies look like.
	 */
	public static final Pattern INTERPOLATE_REGEX = Pattern.compile("\\$\\{ *([a-zA-Z0-9_]+) *\\}");

	/**
	 * Replaces all variable references within <tt>text</tt> with the variable
	 * values in <tt>vars</tt>.
	 * 
	 * @param text Text to replace variables in
	 * @param vars Variable values to replace with
	 * @return Text with variables replace
	 */
	public static String interpolate(String text, VladMessageVars vars) {
        Matcher m = INTERPOLATE_REGEX.matcher(text);
        StringBuffer sb = new StringBuffer();
        // would use m.appendReplacement() and m.appendTail() except that they require escaping
        // of backslash and dollar sign characters
        int appendLocation = 0;
        while (m.find()) {
            String var = m.group(1);
            String val = vars.get(var);
            if (val == null) {
            	val = "";
            	if (log.isDebugEnabled())
            		log.debug("Undefined variable '"+var+"' referenced in: "+text);
            }
            int matchPrior = m.start();
            if (matchPrior - appendLocation > 0)
            	sb.append(text, appendLocation, matchPrior);
            sb.append(val);
            appendLocation = m.end();
        }
        sb.append(text, appendLocation, text.length());
        return sb.toString();
	}
	
	/**
	 * Uses type-inference to make HashMap declarations compact.
	 * You can write <tt>Map<K,V> m = VladUtil.newHashMap()</tt> instead
	 * of the more verbose <tt>Map<K,V> m = new HashMap<K,V>()</tt>  
	 */
	public static <K,V> HashMap<K,V> newHashMap() {
		return new HashMap<K,V>();
	}

	/**
	 * Uses type-inference to make LinkedHashMap declarations compact.
	 * You can write <tt>Map<K,V> m = VladUtil.newLinkedHashMap()</tt> instead
	 * of the more verbose <tt>Map<K,V> m = new LinkedHashMap<K,V>()</tt>  
	 */
	public static <K,V> LinkedHashMap<K,V> newLinkedHashMap() {
		return new LinkedHashMap<K,V>();
	}

	/**
	 * Uses type-inference to make ArrayList declarations compact.
	 * You can write <tt>List<V> l = VladUtil.newArrayList()</tt> instead
	 * of the more verbose <tt>List<V> l = new ArrayList<V>()</tt>  
	 */
	public static <V> ArrayList<V> newArrayList() {
		return new ArrayList<V>();
	}

	/**
	 * Uses type-inference to make HashSet declarations compact.
	 * You can write <tt>Set<V> s = VladUtil.newHashSet()</tt> instead
	 * of the more verbose <tt>Set<V> s = new HashSet<V>()</tt>  
	 */
	public static <V> HashSet<V> newHashSet() {
		return new HashSet<V>();
	}
}
