/* Copyright 2009 Patrick Higgins 
 *
 * This file is part of VladNput.
 *
 * VladNput is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * VladNput is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with VladNput.  If not, see <http://www.gnu.org/licenses/>.
 */
tree grammar VladWalker;

options {
  tokenVocab=VladParser;
  ASTLabelType=CommonTree;
  superClass=VladTreeParser;
}

@header {
package org.vladnput;

import java.util.Map;
import java.util.LinkedHashMap;
import java.util.regex.Pattern;
}

@members {

private final static boolean DEBUG = false;

public void setParseErrors(ParseErrors parseErrors) {
    setErrors(parseErrors);
}

public String getErrorHeader(RecognitionException e) {
    return VladUtil.getSourceLocation(e);
}

public void emitErrorMessage(String msg) {
    getErrors().add(msg);
}

public String getErrorMessage(RecognitionException e, String[] tokenNames) {
    if (DEBUG) {
        List stack = getRuleInvocationStack(e, this.getClass().getName());
        String msg = null;
        if (e instanceof NoViableAltException) {
            NoViableAltException nvae = (NoViableAltException) e;
            msg = " no viable alt; token="+e.token+
                  " (decision="+nvae.decisionNumber+
                  " state "+nvae.stateNumber+")"+
                  " decision=<<"+nvae.grammarDecisionDescription+">>";
        }
        else {
            msg = super.getErrorMessage(e, tokenNames);
        }
        return stack+" "+msg;
    }
    else {
        return super.getErrorMessage(e, tokenNames);
    }
}
    
public String getTokenErrorDisplay(Token t) {
    return DEBUG ? t.toString() : super.getTokenErrorDisplay(t);
}

}

start
	:	stmt+ EOF
	;

stmt
	:	include_stmt
	|	options_stmt
	|	type_stmt
	|	param_stmt
	|	param_match_stmt
	|	url_stmt
	|	url_match_stmt
	;

include_stmt
	:	^(INCLUDE
		  ( string		{ includeFile($string.value, $string.tree); }
		  | ANGLE_STRING	{ includeResource($ANGLE_STRING); }
	          )
	         )
	;

options_stmt
	:	^(OPTIONS option+)
	|	OPTIONS
	;
	
option
	:	^(OPTION ID literal)	{ addOption($ID, $literal.value, $literal.tree); }
	|	msg_defaults
	;

type_stmt
	:	^(TYPE ID type_expr)	{ defineType($ID, $type_expr.value); }
	;
	
param_stmt
@after { finishParam((Param) currentParam); currentParam = null; }
	:	^(PARAM
		  param_id
		  card_spec
		  param_args
		 )
	;
	
param_id
	:	ID		{ currentParam = addParam($ID.text, $ID); }
	|	string	{ currentParam = addParam($string.value, $string.tree); }
	;
	
param_match_stmt
@after { finishParamMatch((ParamMatch) currentParam); currentParam = null; }
	:	^(PARAM_MATCH
		  regex		{ currentParam = addParamMatch($regex.value, $regex.tree); }
		  card_spec
		  param_args
		 )
	;

url_stmt
	:	^(URL
		  url		{ currentURL = addURL($url.value, $url.tree); }
		  url_body
		 )
	;

url_match_stmt
	:	^(URL_MATCH
		  regex		{ currentURL = addURLMatch($regex.value, $regex.tree); }
		  url_body
		 )
	;

url returns [String value, CommonTree tree]
	:	URL_LITERAL	{ $value = $URL_LITERAL.text; $tree = $URL_LITERAL; }
	|	string		{ $value = $string.value; $tree = $string.tree; }
	;

url_body
@after { currentURL = null; }
	:	^(URL_BODY url_body_stmt+)
	|	NO_BODY
	;
	
url_body_stmt
	:	param_stmt
	|	param_match_stmt
	;
	
card_spec
	:	QUESTION	{ currentParam.setCardinality(Cardinality.OPTIONAL); }
	|	STAR		{ currentParam.setCardinality(Cardinality.ZERO_OR_MORE); }
	|	PLUS		{ currentParam.setCardinality(Cardinality.ONE_OR_MORE); }
	|	CARET		{ currentParam.setCardinality(null); }
	|	^(LBRACE min=INT (max=INT)?)
				{ setCardinality(currentParam, $min, $max); }
	|	NO_CARD_SPEC	{ currentParam.setCardinality(Cardinality.REQUIRED); }
	;

param_args
	:	type_expr	{ currentParam.setType($type_expr.value); }
	|	param_kw_args
	|	NO_ARGS
	;
	
type_expr returns [Type value]
	:	( typeName=ID		{ $value = new Type($typeName, resolveType($typeName)); }
		| ^(QNAME (s+=ID)+)	{ $value = new Type((CommonTree) $s.get(0), parseClass($s)); }
		)
		(type_kw_args[$value])?
	;

msg_defaults
	:	^(MESSAGE_DEFAULTS ^(KW_ARGS msg_defaults_kw_arg+))
	;

msg_defaults_kw_arg
	:	^(ID string)	{ addMessageDefault($ID, $string.value); }
	;

param_kw_args
	:	^(KW_ARGS param_kw_arg+ )
	|	EMPTY_ARGS
	;

param_kw_arg
	:	^(TYPE type_expr)	{ currentParam.setType($type_expr.value); }
	|	^(ID literal)		{ addParamKwArg($ID, $literal.value); }
	;

type_kw_args[Type type]
@after { typeArgsComplete(type); }
	:	^(KW_ARGS (type_kw_arg[type])+)
	|	EMPTY_ARGS
	;

type_kw_arg[Type type]
	:	^(ID literal)		{ addTypeArg(type, $ID, $literal.value); }
	;

literal returns [Object value, CommonTree tree]
	:	string	{ $value = $string.value; $tree = $string.tree; }
	|	INT	{ $value = parseInt($INT); $tree = $INT; }
	|	FLOAT	{ $value = parseFloat($FLOAT); $tree = $FLOAT; }
	|	TRUE	{ $value = Boolean.TRUE; $tree = $TRUE; }
	|	FALSE	{ $value = Boolean.FALSE; $tree = $FALSE; }
	;

regex returns [Pattern value, CommonTree tree]
	:	string { $value = compileRegex($string.value, $string.tree); $tree = $string.tree; }
	;

string returns [String value, CommonTree tree]
	:	( v=DOUBLEQUOTED_STRING
	|	| v=SINGLEQUOTED_STRING
		) { $value = parseString($v); $tree = $v; }
	;
