/* Copyright 2009 Patrick Higgins 
 *
 * This file is part of VladNput.
 *
 * VladNput is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * VladNput is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with VladNput.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vladnput.types;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

import org.vladnput.VladMessageVars;

public class CreditCardType extends VladType {

	// not a valid argument, but a message for an invalid luhn can be given (luhn_msg)
	public static final String LUHN_KEY = "luhn";

	public static final String FORMAT_KEY = "format";
	public static final String TYPES_ALLOWED_KEY = "types_allowed";
	
	public static final int VISA             = (1<<0);
	public static final int MASTERCARD       = (1<<1);
	public static final int AMERICAN_EXPRESS = (1<<2);
	public static final int ALL_TYPES        = VISA | MASTERCARD | AMERICAN_EXPRESS;
	
	private static final Pattern VISA_PATTERN = Pattern.compile("4\\d{12}(?:\\d{3})?");
	private static final Pattern MASTERCARD_PATTERN = Pattern.compile("5[1-5]\\d{14}");
	private static final Pattern AMERICAN_EXPRESS_PATTERN = Pattern.compile("3[47]\\d{13}");
	
	public static final Set<String> VALID_KEYS;
	public static final Map<String,Integer> TYPE_CODES;
	static {
		Set<String> validKeys = new HashSet<String>();

		validKeys.add(FORMAT_KEY);
		validKeys.add(TYPES_ALLOWED_KEY);

		VALID_KEYS = Collections.unmodifiableSet(validKeys);

		VladMessageVars.registerVars(VALID_KEYS);
		
		Map<String,Integer> typeCodes = new HashMap<String,Integer>();
		typeCodes.put("visa", Integer.valueOf(VISA));
		typeCodes.put("mastercard", Integer.valueOf(MASTERCARD));
		typeCodes.put("master_card", Integer.valueOf(MASTERCARD));
		typeCodes.put("mc", Integer.valueOf(MASTERCARD));
		typeCodes.put("american_express", Integer.valueOf(AMERICAN_EXPRESS));
		typeCodes.put("amex", Integer.valueOf(AMERICAN_EXPRESS));
		typeCodes.put("all", Integer.valueOf(ALL_TYPES));
		TYPE_CODES = Collections.unmodifiableMap(typeCodes);
	}
	
	private final Pattern pattern;
	private final int typesAllowed;

	public CreditCardType(Map<String,Object> args) {
		super(args);
		this.pattern = Pattern.compile(stringKey(args, FORMAT_KEY));
		this.typesAllowed = getTypesAllowed(stringKey(args, TYPES_ALLOWED_KEY));
	}
	
	private static int getTypesAllowed(String value) throws IllegalArgumentException {
		if (value == null)
			return ALL_TYPES;

		String values[] = value.split("[, ]+");
		int types = 0;
		for (int i = 0; i < values.length; i++) {
			String cardName = values[i].toLowerCase();
			Integer code = (Integer) TYPE_CODES.get(cardName);
			if (code == null)
				throw new IllegalArgumentException(value);
			types |= code.intValue();
		}		
		return types;		
	}
	
	public static void checkArgs(Map<String,Object> args) throws MissingArgException {
		if (!args.containsKey(FORMAT_KEY))
			throw new MissingArgException(FORMAT_KEY);
	}

	public static void checkArg(String name, Object value, Map<String,Object> args) throws InvalidArgException, InvalidArgValueException {
		if (VALID_KEYS.contains(name)) {
			if (FORMAT_KEY.equals(name)) {
				try {
					Pattern.compile((String) value);
				}
				catch (PatternSyntaxException e) {
					throw new InvalidArgValueException(name, value);
				}
			}
			else if (TYPES_ALLOWED_KEY.equals(name)) {
				try {
					getTypesAllowed((String) value);
				}
				catch (IllegalArgumentException e) {
					throw new InvalidArgValueException(name, value);
				}
			}
		}
		else {
			VladType.checkArg(name, value, args);
		}
	}
	
	private static boolean validLuhn(String cardNumber) {
		int sum = 0;
		int digit = 0;
		int addend = 0;
	    boolean timesTwo = false;
	    char[] chars = cardNumber.toCharArray();

	    for (int i = chars.length - 1; i >= 0; i--) {
		    digit = chars[i] - '0';
		    if (timesTwo) {
		    	addend = digit * 2;
		    	if (addend > 9) {
		    		addend -= 9;
		    	}
		    }
		    else {
		    	addend = digit;
		    }
		    sum += addend;
		    timesTwo = !timesTwo;
	    }

	    return (sum % 10) == 0;
	}
	
	private boolean validType(String cardNumber) {
		return
			(typesAllowed & VISA) != 0 && VISA_PATTERN.matcher(cardNumber).matches() ||
			(typesAllowed & MASTERCARD) != 0 && MASTERCARD_PATTERN.matcher(cardNumber).matches() ||
			(typesAllowed & AMERICAN_EXPRESS) != 0 && AMERICAN_EXPRESS_PATTERN.matcher(cardNumber).matches();
	}

	@Override
	protected List<String> doValidate(List<String> msgs, String value, VladMessageVars vars) {
		msgs = super.doValidate(msgs, value, vars);
		Matcher matcher = pattern.matcher(value);
		if (!matcher.matches()) {
			msgs = addMessage(msgs, FORMAT_KEY);
			vars.put(FORMAT_KEY, pattern.pattern());
		}
		else {
			String cardNumber = value.replaceAll("[^0-9]+", "");
			// valid format, now check luhn and types
			if (!validLuhn(cardNumber)) {
				msgs = addMessage(msgs, LUHN_KEY);
			}
			else if (!validType(cardNumber)) {
				msgs = addMessage(msgs, TYPES_ALLOWED_KEY);
			}
		}
		return msgs;
	}

	/**
	 * Returns true because credit cards are always sensitive.
	 */
	public boolean isSensitive() {
		return true;
	}

}
