/* Copyright 2009 Patrick Higgins 
 *
 * This file is part of VladNput.
 *
 * VladNput is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * VladNput is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with VladNput.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vladnput.types;

import java.text.DateFormat;
import java.text.ParseException;
import java.text.ParsePosition;
import java.text.SimpleDateFormat;
import java.util.Collections;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.vladnput.VladMessageVars;
import org.vladnput.VladMessages;
import org.vladnput.VladUtil;

public class DateType extends VladType {

	public static final String FORMAT_KEY = "format";
	public static final String LATEST_ALLOWED_KEY = "latest_allowed";
	public static final String EARLIEST_ALLOWED_KEY = "earliest_allowed";
	
	public static final Set<String> VALID_KEYS;
	static {
		Set<String> validKeys = new HashSet<String>();

		validKeys.add(FORMAT_KEY);
		validKeys.add(LATEST_ALLOWED_KEY);
		validKeys.add(EARLIEST_ALLOWED_KEY);

		VALID_KEYS = Collections.unmodifiableSet(validKeys);

		VladMessageVars.registerVars(VALID_KEYS);
	}
	
	private final String formatString;
	private final DateFormat format;
	private final Date earliestAllowed;
	private final Date latestAllowed;

	public DateType(Map<String,Object> args) {
		super(args);
		this.formatString = getFormatString(stringKey(args, FORMAT_KEY));
		this.format = getFormat(formatString);
		this.earliestAllowed = dateKey(args, EARLIEST_ALLOWED_KEY, format);
		this.latestAllowed = dateKey(args, LATEST_ALLOWED_KEY, format);
	}
	
	private static String getFormatString(String fmt) {
		if (fmt == null)
			fmt = "MM/dd/yyyy";
		return fmt;
	}
	
	private static DateFormat getFormat(String fmt) {
		fmt = getFormatString(fmt);
		DateFormat format = new SimpleDateFormat(fmt);
		format.setLenient(false);
		return format;
	}

	public static void checkArg(String name, Object value, Map<String,Object> args) throws InvalidArgException, InvalidArgValueException {
		if (VALID_KEYS.contains(name)) {
			if (FORMAT_KEY.equals(name)) {
				try {
					getFormat((String) value);
				}
				catch (IllegalArgumentException e) {
					throw new InvalidArgValueException(name, value);
				}
			}
			else {
				try {
					Date date = VladUtil.parseDate((String) value, getFormat((String) args.get(FORMAT_KEY)));
					if (LATEST_ALLOWED_KEY.equals(name)) {
						Object minObj = (Object) args.get(EARLIEST_ALLOWED_KEY);
						if (minObj instanceof String) { // don't cause ClassCastException for earliest allowed
							try {
								Date minDate = VladUtil.parseDate((String) value, getFormat((String) minObj));
								if (date.before(minDate))
									throw new InvalidArgValueException(name, value);
							}
							catch (IllegalArgumentException e) {
								// don't worry about parse errors for earliest allowed
							}
						}
					}
				}
				catch (IllegalArgumentException e) {
					throw new InvalidArgValueException(name, value);
				}
			}
		}
		else {
			VladType.checkArg(name, value, args);
		}
	}
	
	@Override
	public List<String> doValidate(List<String> msgs, String value, VladMessageVars vars) {
		msgs = super.doValidate(msgs, value, vars);
		if (value.length() > formatString.length()) {
			msgs = addMessage(msgs, FORMAT_KEY);
			addVars(vars);			
		}
		else {
			ParsePosition pos = new ParsePosition(0);
			Date d = format.parse(value, pos);
			if (d == null || pos.getIndex() != value.length()) {
				msgs = addMessage(msgs, FORMAT_KEY);
				addVars(vars);
			}
			else if (d != null) {
				if (earliestAllowed != null && d.before(earliestAllowed)) {
					msgs = addMessage(msgs, EARLIEST_ALLOWED_KEY);
					addVars(vars);					
				}
				if (latestAllowed != null && d.after(latestAllowed)) {
					msgs = addMessage(msgs, LATEST_ALLOWED_KEY);
					addVars(vars);					
				}
			}
		}
		return msgs;
	}
	
	private void addVars(VladMessageVars vars) {
		vars.put(FORMAT_KEY, formatString);
		vars.put(EARLIEST_ALLOWED_KEY, earliestAllowed == null ? "" : format.format(earliestAllowed));
		vars.put(LATEST_ALLOWED_KEY, latestAllowed == null ? "" : format.format(latestAllowed));
	}

}
