/* Copyright 2009 Patrick Higgins 
 *
 * This file is part of VladNput.
 *
 * VladNput is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * VladNput is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with VladNput.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vladnput.types;

import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

import org.vladnput.VladMessageVars;

public class StringType extends VladType {
	
	public static final String REGEX_KEY = "regex";
	public static final String CASE_INSENSITIVE_KEY = "case_insensitive";
	public static final String PARTIAL_KEY = "partial";
	static {
		VladMessageVars.registerVar(REGEX_KEY);
		VladMessageVars.registerVar(CASE_INSENSITIVE_KEY);
		VladMessageVars.registerVar(PARTIAL_KEY);
	}

	private final Pattern pattern;
	private final boolean partial;

	public StringType(Map<String,Object> args) {
		super(args);
		int flags = 0;
		String patternString = stringKey(args, REGEX_KEY);
		if (patternString != null) {
			Boolean caseInsensitive = booleanKey(args, CASE_INSENSITIVE_KEY);
			if (caseInsensitive != null && Boolean.TRUE.equals(caseInsensitive)) {
				flags |= Pattern.CASE_INSENSITIVE;
			}
			this.pattern = Pattern.compile(patternString, flags);
		}
		else {
			this.pattern = null;
		}
		this.partial = Boolean.TRUE.equals(booleanKey(args, PARTIAL_KEY));
	}

	public static void checkArgs(Map<String,Object> args) throws MissingArgException {
		if (!args.containsKey(REGEX_KEY))
			throw new MissingArgException(REGEX_KEY);
	}

	public static void checkArg(String name, Object value, Map<String,Object> args) throws InvalidArgException, InvalidArgValueException {
		if (REGEX_KEY.equals(name)) {
			try {
				Pattern.compile((String) value);
			}
			catch (PatternSyntaxException e) {
				throw new InvalidArgValueException(name, value);
			}
		}
		else if (CASE_INSENSITIVE_KEY.equals(name)) {
			((Boolean) value).booleanValue(); // the optimizer better not remove this code!
		}
		else if (PARTIAL_KEY.equals(name)) {
			((Boolean) value).booleanValue(); // the optimizer better not remove this code!
		}
		else if (ALLOW_EMPTY_KEY.equals(name)) {
			((Boolean) value).booleanValue(); // the optimizer better not remove this code!
		}
		else {
			VladType.checkArg(name, value, args);
		}
	}
	
	public Pattern getPattern() {
		return pattern;
	}
	
	public boolean isPartial() {
		return partial;
	}

	@Override
	public List<String> doValidate(List<String> msgs, String value, VladMessageVars vars) {
		msgs = super.doValidate(msgs, value, vars);
		Matcher matcher = pattern.matcher(value);
		if (! (partial ? matcher.find() : matcher.matches())) {
			msgs = addMessage(msgs, REGEX_KEY);
			vars.put(REGEX_KEY, pattern.pattern());
			if ((pattern.flags() & Pattern.CASE_INSENSITIVE) == Pattern.CASE_INSENSITIVE)
				vars.put(CASE_INSENSITIVE_KEY, "true");
			else
				vars.put(CASE_INSENSITIVE_KEY, "false");
			vars.put(PARTIAL_KEY, ""+partial);
		}
		return msgs;
	}

}
