/* Copyright 2009 Patrick Higgins 
 *
 * This file is part of VladNput.
 *
 * VladNput is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * VladNput is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with VladNput.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vladnput.types;

import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

import org.vladnput.VladMessageVars;

public class URLType extends VladType {
	
	public static final String RELATIVE_KEY = "relative";

	// the following regexps are compiled with Pattern.COMMENTS (same as Perl //x modifier)
	// so that whitespace can be inserted to make them a little easier to read
	private static final String RELATIVE_REGEX =
		"(?!https?://)" +                             // don't allow full URLs
		"(?: [/\\p{Alnum}][\\p{Graph}&&[^?\\#]]* )" + // path info, required
		"(?: \\? [\\p{Graph}&&[^?\\#]]+ )?" +         // query string
		"(?: \\# [\\p{Graph}&&[^?\\#]]+ )?";          // anchor

	private static final String FULL_REGEX =
		"https?://" +                                 // scheme
		"[-a-zA-Z0-9]+ (?: \\.[-a-zA-Z0-9]+ )*" +     // host name
		"(?: :\\d+ )?" +                              // port
		"/" +                                         // trailing slash is required
		"(?: [\\p{Alnum}][\\p{Graph}&&[^?\\#]]* )?" + // path info is optional because the trailing slash is required
		"(?: \\? [\\p{Graph}&&[^?\\#]]+ )?" +         // query string
		"(?: \\# [\\p{Graph}&&[^?\\#]]+ )?";          // anchor

	private static final Pattern FULL_PATTERN =
		Pattern.compile(FULL_REGEX, Pattern.COMMENTS | Pattern.CASE_INSENSITIVE);
	private static final Pattern RELATIVE_PATTERN =
		Pattern.compile(RELATIVE_REGEX, Pattern.COMMENTS | Pattern.CASE_INSENSITIVE);
	
	private final Pattern pattern;

	public URLType(Map<String,Object> args) {
		super(args);
		pattern = Boolean.TRUE.equals(booleanKey(args, RELATIVE_KEY)) ?
			RELATIVE_PATTERN : FULL_PATTERN;
	}
	
	public static void checkArg(String name, Object value, Map<String,Object> args) throws InvalidArgException, InvalidArgValueException {
		if (RELATIVE_KEY.equals(name)) {
			((Boolean) value).booleanValue(); // the optimizer better not remove this code!
		}
		else {
			VladType.checkArg(name, value, args);
		}
	}
	
	public boolean isRelative() {
		return pattern == RELATIVE_PATTERN;
	}
	
	@Override
	public List<String> doValidate(List<String> msgs, String value, VladMessageVars vars) {
		msgs = super.doValidate(msgs, value, vars);
		if (!pattern.matcher(value).matches()) {
			msgs = addMessage(msgs, "url");
		}
		return msgs;
	}

}
