/* Copyright 2009 Patrick Higgins 
 *
 * This file is part of VladNput.
 *
 * VladNput is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * VladNput is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with VladNput.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vladnput.types;

import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.vladnput.VladMessageVars;
import org.vladnput.VladMessages;
import org.vladnput.VladUtil;

/**
 * The base class for all type validators. Typically, a subclass will define
 * arguments that it handles as string constants and check them in 
 *
 * @author Patrick Higgins
 */
abstract public class VladType {

	/**
	 * Message variable for the length of the actual value.
	 */
	public static final String LEN_VAR = "len";
	/**
	 * Name of the minimum required length argument. Also, name of the message variable for the minimum required length.
	 */
	public static final String MIN_LEN_KEY = "min_len";
	/**
	 * Name of the maximum allowed length argument. Also, name of the message variable for the maximum allowed length.
	 */
	public static final String MAX_LEN_KEY = "max_len";
	/**
	 * Name of the sensitive argument.
	 */
	public static final String SENSITIVE_KEY = "sensitive";
	/**
	 * Name of the allow empty argument.
	 */
	public static final String ALLOW_EMPTY_KEY = "allow_empty";

	static {
		VladMessageVars.registerVar(LEN_VAR);
		VladMessageVars.registerVar(MIN_LEN_KEY);
		VladMessageVars.registerVar(MAX_LEN_KEY);
		VladMessageVars.registerVar(ALLOW_EMPTY_KEY);
	}
	
	/**
	 * Set of valid argument names
	 */
	public static final Set<String> VALID_KEYS;
	static {
		Set<String> validKeys = new HashSet<String>();

		validKeys.add(MIN_LEN_KEY);
		validKeys.add(MAX_LEN_KEY);
		validKeys.add(SENSITIVE_KEY);
		validKeys.add(ALLOW_EMPTY_KEY);

		VALID_KEYS = Collections.unmodifiableSet(validKeys);
	}

	/**
	 * Checks argument with <tt>name</tt> and <tt>value</tt> for validity. All
	 * other arguments for this type are given in <tt>args</tt> so that
	 * dependencies between arguments can be checked, for example that min_len
	 * is less than max_len.
	 * 
	 * Subclasses will likely need to "override" this method, which seems
	 * confusing because this method is static and thus cannot be overridden.
	 * The method is called through reflection and thus defining a static
	 * method with the same name and signature in a subclass will cause the
	 * subclass method to be called. For examples, see <tt>StringType</tt>
	 * and <tt>DateType</tt>.
	 *  
	 * @param name Argument name
	 * @param value Argument value
	 * @param args The other arguments for this type
	 * @throws InvalidArgException If the argument is invalid
	 * @throws InvalidArgValueException If the argument value is invalid
	 */
	public static void checkArg(String name, Object value, Map<String,Object> args) throws InvalidArgException, InvalidArgValueException {
		if (VladMessages.isValidMessageName(name)) {
			if (!VladMessages.isValidFormat((String) value))
				throw new InvalidArgValueException(name, value);
		}
		else if (VALID_KEYS.contains(name)) {
			if (SENSITIVE_KEY.equals(name)) {
				((Boolean) value).booleanValue(); // an optimizer better not throw this out!
			}
			else if (ALLOW_EMPTY_KEY.equals(name)) {
				((Boolean) value).booleanValue(); // an optimizer better not throw this out!				
			}
			else {
				int len = ((Integer) value).intValue();
				if (len < -1)
					throw new InvalidArgValueException(name, value);
				if (MAX_LEN_KEY.equals(name)) {
					Object minObj = (Object) args.get(MIN_LEN_KEY);
					if (minObj instanceof Integer) { // don't cause ClassCastException for min_len
						int minLen = ((Integer) minObj).intValue();
						if (len < minLen)
							throw new InvalidArgValueException(name, value);
					}
				}
			}
		}
		else {
			throw new InvalidArgException(name);
		}
	}

	/**
	 * Returns the name of the message argument for <tt>arg</tt>. This will
	 * always be <tt>arg</tt> with "_msg" appended. 
	 * @param arg Argument name
	 * @return Name of corresponding message argument
	 */
	public static String msgFor(String arg) {
		return arg+"_msg";
	}

	private final VladMessages messages;
	private final Integer minLen;
	private final Integer maxLen;
	private final boolean sensitive;
	private final boolean allowEmpty;

	/**
	 * Constructs VladType from <tt>args</tt>, removing any arguments related to
	 * <tt>VladMessages</tt> or that are recognized by this class (e.g. min_len
	 * and max_len).
	 * 
	 * Any arguments not recognized are ignored.
	 * 
	 * @param args Arguments for this type
	 */
	public VladType(Map<String,Object> args) {
		this.messages = new VladMessages(args);
		this.minLen = integerKey(args, MIN_LEN_KEY);
		this.maxLen = integerKey(args, MAX_LEN_KEY);
		this.sensitive = Boolean.TRUE.equals(booleanKey(args, SENSITIVE_KEY));
		this.allowEmpty = Boolean.TRUE.equals(booleanKey(args, ALLOW_EMPTY_KEY));
	}

	/**
	 * Adds the name of the message for argument <tt>name</tt> to <tt>msgs</tt>.
	 * If a <tt>null</tt> list is passed for <tt>msgs</tt>, a new list is
	 * constructed and returned with the message name added to it.  
	 * @param msgs List of messages being accumulated
	 * @param name Argument name to add a message for
	 * @return List of message names. Will either be same as <tt>msgs</tt> or a new list if <tt>msgs</tt> is <tt>null</tt>.
	 */
	public List<String> addMessage(List<String> msgs, String name) {
		if (msgs == null)
			msgs = new ArrayList<String>();
		msgs.add(msgFor(name));
		return msgs;
	}
	
	/**
	 * Gets the messages for this type.
	 * @return Messages for this type
	 */
	public VladMessages getMessages() {
		return messages;
	}
	
	/**
	 * Checks <tt>value</tt> for validity. If it is valid, a <tt>null</tt>
	 * value will be returned. If it is invalid, a list of message names
	 * describing the error will be returned and the variables that may be
	 * used by those messages will be placed into <tt>vars</tt>.
	 * 
	 * @param value Value to validate
	 * @param vars Variables to add values to if the value is invalid
	 * @return <tt>null</tt> if the value is valid, a list of message names otherwise
	 */
	public final List<String> validate(String value, VladMessageVars vars) {
		List<String> msgs = null;		
		if (allowEmpty() && value.length() == 0) {
			return msgs;
		}
		return doValidate(msgs, value, vars);
	}
	
	protected List<String> doValidate(List<String> msgs, String value, VladMessageVars vars) {
		int len = value.length();
		if (minLen != null && len < minLen.intValue()) {
			msgs = addMessage(msgs, MIN_LEN_KEY);
			vars.put(MIN_LEN_KEY, minLen.toString());
			vars.put(LEN_VAR, Integer.toString(len));
		}
		if (maxLen != null && len > maxLen.intValue()) {
			msgs = addMessage(msgs, MAX_LEN_KEY);
			vars.put(MAX_LEN_KEY, maxLen.toString());
			vars.put(LEN_VAR, Integer.toString(len));
		}
		return msgs;
	}

	public boolean isSensitive() {
		return sensitive;
	}

	public boolean allowEmpty() {
		return allowEmpty;
	}
	
	/**
	 * Removes and returns the mapping associated with <tt>key</tt> in
	 * <tt>args</tt>, or <tt>null</tt> if there is not one.
	 * @param args Arguments
	 * @param key Argument name
	 * @return Argument value, or <tt>null</tt> if there is not one.
	 */
	protected String stringKey(Map<String,Object> args, String key) {
		return VladUtil.stringKey(args, key);
	}
	
	/**
	 * Removes and returns the mapping associated with <tt>key</tt> in
	 * <tt>args</tt>, or <tt>null</tt> if there is not one.
	 * @param args Arguments
	 * @param key Argument name
	 * @return Argument value, or <tt>null</tt> if there is not one.
	 */
	protected Integer integerKey(Map<String,Object> args, String key) {
		return VladUtil.integerKey(args, key);
	}

	/**
	 * Removes and returns the mapping associated with <tt>key</tt> in
	 * <tt>args</tt>, or <tt>null</tt> if there is not one.
	 * @param args Arguments
	 * @param key Argument name
	 * @return Argument value, or <tt>null</tt> if there is not one.
	 */
	protected Double floatKey(Map<String,Object> args, String key) {
		return VladUtil.floatKey(args, key);
	}

	/**
	 * Removes and returns the mapping associated with <tt>key</tt> in
	 * <tt>args</tt>, or <tt>null</tt> if there is not one.
	 * @param args Arguments
	 * @param key Argument name
	 * @return Argument value, or <tt>null</tt> if there is not one.
	 */
	protected Boolean booleanKey(Map<String,Object> args, String key) {
		return VladUtil.booleanKey(args, key);
	}

	/**
	 * Removes and returns the string mapping associated with <tt>key</tt> in
	 * <tt>args</tt> after parsing it as a date, or <tt>null</tt> if there is
	 * not one.
	 * @param args Arguments
	 * @param key Argument name
	 * @param format Date format
	 * @return Argument value, or <tt>null</tt> if there is not one.
	 * @throws IllegalArgumentException If the date is invalid
	 */
	protected Date dateKey(Map<String,Object> args, String key, DateFormat format) throws IllegalArgumentException {
		return VladUtil.dateKey(args, key, format);
	}

	public String toString() {
		StringBuffer b = new StringBuffer();
		Class<?> klass = this.getClass();
		String className = klass.getName();
		b.append(className).append("(");
		while (klass != null) {
			Field[] fields = klass.getDeclaredFields();
			for (int i = 0; i < fields.length; i++) {
				try {
					if (!Modifier.isStatic(fields[i].getModifiers())) {
						fields[i].setAccessible(true);
						b.append(fields[i].getName()).append("=").append(fields[i].get(this)).append(",");
					}
				} catch (IllegalArgumentException e) {
					throw new RuntimeException(e);
				} catch (IllegalAccessException e) {
					throw new RuntimeException(e);
				}
			}
			klass = klass.getSuperclass();
		}
		b.append(")");
		return b.toString();
	}

}
