/* Copyright 2009 Patrick Higgins 
 *
 * This file is part of VladNput.
 *
 * VladNput is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * VladNput is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with VladNput.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vladnput;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URL;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;

import javax.servlet.http.HttpServletRequest;

class RequestLogParser implements Iterable<HttpServletRequest> {

	private final BufferedReader reader;
	private final File file;
	private int lineNumber = 0;
	
	public RequestLogParser(URL url) throws IOException, RequestLogParseException {
		this(new File(url.getPath()));
	}

	public RequestLogParser(File file) throws IOException, RequestLogParseException {
		this.file = file;
		this.reader = new BufferedReader(new FileReader(file));
		this.lineNumber = 0;
	}
	
	public Iterator<HttpServletRequest> iterator() {
		try {
			return new LogEntryIterator();
		}
		catch (IOException e) {
			throw new RuntimeException(e);
		}
	}
	
	public void close() {
		if (reader != null) {
			try {
				reader.close();
			}
			catch (IOException ignore) { }
		}
	}

	private static String decode(String s) {
		try {
			return URLDecoder.decode(s, "UTF-8");
		}
		catch (UnsupportedEncodingException e) {
			throw new RuntimeException(e);
		}
	}
	
	private class LogEntryIterator implements Iterator<HttpServletRequest> {

		private LogEntry logEntry;
		
		LogEntryIterator() throws RequestLogParseException, IOException {
			this.logEntry = readNext();
		}
	
		private synchronized LogEntry readNext() throws RequestLogParseException, IOException {
			final int MARK_READAHEAD = 8192;
			LogEntry entry = null;
			String line = null;
			while (true) {
				reader.mark(MARK_READAHEAD);
				line = reader.readLine();
				if (line == null) {
					break;
				}

				++lineNumber;
				if (line.trim().length() == 0)
					continue; // skip blank lines

				if (line.charAt(0) == '#') // skip comments
					continue;
				
				if (line.charAt(0) == '\t') {
					if (entry == null) {
						throw new RequestLogParseException(file, lineNumber, "params without entry");
					}
					String[] vals = line.split("[=\\t]");
					if (vals.length == 3) {
						entry.addParam(decode(vals[1]), decode(vals[2]));
					}
					else if (vals.length == 2) {
						entry.addParam(decode(vals[1]), "");
					}
					else if (line.length() == 2) {
						// allow "\t=" lines. Don't know what these are, but we have them...
					}
					else {
						throw new RequestLogParseException(file, lineNumber, "invalid line format");
					}
				}
				else {
					if (entry != null) {
						reader.reset();
						--lineNumber;
						break;
					}
					String[] vals = line.split(" ", 5);
					if (vals.length != 5) {
						throw new RequestLogParseException(file, lineNumber, "not enough values");
					}
					String remoteAddr = vals[2];
					String method = vals[3];
					String url = decode(vals[4]);
					entry = new LogEntry(method, url, remoteAddr, lineNumber);
				}
			}
			return entry;
		}
		
		public boolean hasNext() {
			return this.logEntry != null;
		}
		
		public LogEntry next() {
			LogEntry entry = this.logEntry;
			try {
				this.logEntry = readNext();
			}
			catch (IOException e) {
				this.logEntry = null;
				throw new RuntimeException(e);
			}
			return entry;
		}

		public void remove() {
			throw new UnsupportedOperationException();
		}
		
	}
	
	static class LogEntry extends StubHttpServletRequest {

		private final String remoteAddr;
		private final int lineNumber;
		
		public LogEntry(String method, String url, String remoteAddr, int lineNumber) {
			super(method, url);
			this.remoteAddr = remoteAddr;
			this.lineNumber = lineNumber;
		}
		
		void addParam(String name, String value) {
			super.addParam(name, value);
		}
		
		public int getLineNumber() {
			return lineNumber;
		}
		
		public List<HttpServletRequest> split() {
			List<HttpServletRequest> list = new ArrayList<HttpServletRequest>();
			for (Enumeration e = getParameterNames(); e.hasMoreElements(); ) {
				String param = (String) e.nextElement(); 
				String[] values = getParameterValues(param);
				for (int i = 0; i < values.length; i++) {
					LogEntry entry = new LogEntry(getMethod(), getRequestURI(), remoteAddr, lineNumber);
					entry.addParam(param, values[i]);
					list.add(entry);
				}
			}
			return list;
		}

		public String getRemoteAddr() {
			return remoteAddr;
		}

		public String getRemoteHost() {
			return remoteAddr;
		}
		
		public String toString() {
			StringBuffer b = new StringBuffer();
			b.append(getLineNumber()).append(":").append(getRequestURI()).append("?");
			for (Enumeration e = getParameterNames(); e.hasMoreElements(); ) {
				String param = (String) e.nextElement(); 
				String[] values = getParameterValues(param);
				for (int i = 0; i < values.length; i++) {
					b.append(param).append("=").append(values[i]).append("&");
				}
			}
			if (b.charAt(b.length()-1) == '&')
				b.setLength(b.length()-1);
			return b.toString();
		}

	}

}
