/* Copyright 2009 Patrick Higgins 
 *
 * This file is part of VladNput.
 *
 * VladNput is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * VladNput is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with VladNput.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vladnput;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringWriter;
import java.util.Iterator;
import java.util.List;

import javax.servlet.http.HttpServletRequest;

import junit.framework.AssertionFailedError;
import junit.framework.TestCase;

import org.vladnput.RequestLogParser.LogEntry;

public class TestUtil {
	
	private TestUtil() { }
	
	public static void logCompare(String[] expected, String log) {
		try {
			String[] lines = log.split("\\r?\\n");
			TestCase.assertEquals(expected.length, lines.length);
			for (int i = 0; i < expected.length; i++) {
				TestCase.assertEquals(expected[i], lines[i]);
			}
		}
		catch (AssertionFailedError e) {
			System.err.println("Expected log:\n");
			for (int i = 0; i < expected.length; i++) {
				System.err.println(expected[i]);
			}
			System.err.println("Actual log:\n"+log);
			throw e;
		}
	}

	public static void logCompare(String expectedResource, String log) throws IOException {
		String expectedLog = null;
		ClassLoader loader = TestUtil.class.getClassLoader();
		Reader reader = null;
		try {
			InputStream input = loader.getResourceAsStream(expectedResource);
			if (input == null)
				throw new RuntimeException("Resource not found: "+expectedResource);
			reader = new InputStreamReader(input);
			StringBuffer buf = new StringBuffer();
			char[] cbuf = new char[8192];
			int nread;
			while ((nread = reader.read(cbuf)) != -1) {
				buf.append(cbuf, 0, nread);
			}
			expectedLog = buf.toString();
		}
		finally {
			if (reader != null)
				reader.close();
		}
		
		try {
			TestCase.assertEquals(expectedLog, log);
		}
		catch (AssertionFailedError e) {
			System.err.println("Expected log:\n"+expectedLog);
			System.err.println("Actual log:\n"+log);
			throw e;
		}
	}
	
	public static void logCompare(String vladFile, String requestFile, String expectLog) throws Exception {
		VladConfig config = VladConfig.parseResource(vladFile);
		ClassLoader loader = TestUtil.class.getClassLoader();
		RequestLogParser parser = new RequestLogParser(loader.getResource(requestFile));
		StringWriter sw = new StringWriter();
		for (HttpServletRequest request : parser) {
			VladErrors errors = config.validate(request, null);
			for (VladError error : errors) {
				sw.write(((LogEntry) request).getLineNumber()+": "+error.getMessage()+"\n");
				List<String> details = error.getDetails();
				if (details != null) {
					for (String detail : details) {
						sw.write("\t"+detail+"\n");
					}
				}
			}
		}
		logCompare(expectLog, sw.toString());
	}

	public static void typeCheck(String vladFile, String requestFile) throws Exception {
		VladConfig config = VladConfig.parseResource(vladFile);
		ClassLoader loader = TestUtil.class.getClassLoader();
		RequestLogParser parser = new RequestLogParser(loader.getResource(requestFile));
		StringWriter sw = new StringWriter();
		for (HttpServletRequest aggregateRequest : parser) {
			boolean shouldFail = aggregateRequest.getRequestURI().contains("fail");
			for (HttpServletRequest request : ((LogEntry) aggregateRequest).split()) {
				VladErrors errors = config.validate(request, null);
				int errorCount = errors.getErrors().size();
				if (shouldFail) {
					if (errorCount == 0) {
						sw.write(((LogEntry) request).getLineNumber()+": fail case passed: "+request+"\n");
					}
					else if (errorCount > 1) {
						sw.write(((LogEntry) request).getLineNumber()+": fail case had too many errors: "+request+", "+errors+"\n");
					}
				}
				else if (errorCount > 0) { // should not have failed, but did
					for (VladError error : errors) {
						sw.write(((LogEntry) request).getLineNumber()+": pass case failed: "+error.getMessage()+"\n");
						List<String> details = error.getDetails();
						if (details != null) {
							for (String detail : details) {
								sw.write("\t"+detail+"\n");
							}
						}
					}
				}
			}
		}
		String s = sw.toString();
		if (s.length() > 0) {
			System.err.println(s);
			throw new AssertionFailedError(s);
		}
	}

}
